#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Sep 30 11:41:27 2020

@author: Peeta Li
@contact: peetal@uoregon.edu
@modulated from https://brainiak.org/tutorials/13-real-time/
"""

# Check pick ~400 brain voxels, from two regions of interest (two cubes). 
# Not choose voxels randomly due to visualization purpose. 
"""
# --------------------------------------------------
# Cherry pick brain voxels for the simulation data  
# --------------------------------------------------
# mPFC MNI cooridnates: (0,54,-6)
# precuneus MNI coordinates: (0,-60,33)

# given MNI coordination, return voxel coordination given a mask dimenison. 
def _MNI_to_voxel_coord(x,y,z,affine):
    import numpy as np
    mni = np.asmatrix([[x],[y],[z],[1]])
    # pusedoinversed affine matrix %*% mni is the voxel coordinate
    voxel = np.asarray(np.asmatrix(np.linalg.pinv(affine.affine)) * mni)
    # transpose the result matrix, choose the first index, round, retrun a list
    center_voxel = [int(c) for c in np.transpose(voxel[0:3])[0].tolist()]
    
    return center_voxel[0], center_voxel[1], center_voxel[2]

# given center voxel coordination, mask and radius, output a roi mask
def _generate_roi_cube(x,y,z,brain_mask,radius):
    # empty the mask
    empty_mask = np.zeros(brain_mask.shape)
    # 
    empty_mask[x-radius:x+radius, y-radius:y+radius, z-radius:z+radius] = 1 
    
    return empty_mask

# load in the mask, can be generated of different sizes. 
dat_dir = '/path/to/FCMA_demo/sim_info'
mask_nii = image.load_img(os.path.join(dat_dir, 'brain_mask.nii.gz'))
mask = mask_nii.get_fdata()

# roi_A (mPFC)
x1, y1, z1 = _MNI_to_voxel_coord(0,54,-6,mask_nii) 
roi_A_mask = _generate_roi_cube(x1,y1,z1,mask,3)
roi_A_mask_nii = image.new_img_like(mask_nii, roi_A_mask, mask_nii.affine)

# roi_B (precuneus)
x2, y2, z2 = _MNI_to_voxel_coord(0,-60,33,mask_nii) 
roi_B_mask = _generate_roi_cube(x2,y2,z2,mask,3)
roi_B_mask_nii = image.new_img_like(mask_nii, roi_B_mask, mask_nii.affine)

# full mask
full_mask = np.zeros(mask.shape)
full_mask[np.where(roi_A_mask == 1)] = 1
full_mask[np.where(roi_B_mask == 1)] = 1
full_mask_nii = image.new_img_like(mask_nii, full_mask, mask_nii.affine)

# write out all the nifts: 
roi_A_mask_nii.to_filename(os.path.join(dat_dir, 'ROI_A_mpfc.nii.gz'))
roi_B_mask_nii.to_filename(os.path.join(dat_dir, 'ROI_B_precuenus.nii.gz'))
full_mask_nii.to_filename(os.path.join(dat_dir, 'full_mask.nii.gz'))
"""

# simulate 10 subjects' fMRI data over 9 small features (ROI) for FCMA demo

import os, random
from nilearn import image, plotting
import numpy as np  
from brainiak.utils import fmrisim as sim 

# specify directory
dat_dir = '/path/to/FCMA_demo/sim_info'

# Specify the volume parameters
trDuration = 1  # seconds
numTRs = 400 # How many TRs will you generate?

# Set up stimulus event time course parameters
event_duration = 15  # How long is each event
isi = 5  # What is the time between each event

# Specify signal magnitude parameters
signal_change = 10 # How much change is there in intensity for the max of the patterns across participants
multivariate_pattern = 1  # Do you want the signal to be a z scored pattern across voxels (1) or a univariate increase (0)

print('Load template of average voxel value')
sub_template_nii = image.load_img(os.path.join(dat_dir, 'brain_template.nii.gz'))
sub_template = sub_template_nii.get_fdata()
dimensions = np.array(sub_template.shape[0:3])

print('Create binary mask and normalize the template range')
mask, template = sim.mask_brain(volume = sub_template, mask_self = True)
mask_cherry = image.load_img(os.path.join(dat_dir, 'full_mask.nii.gz')).get_fdata()

# Load the noise dictionary
print('Loading noise parameters')
with open(os.path.join(dat_dir, 'sub_noise_dict.txt'), 'r') as f:
    noise_dict = f.read()
noise_dict = eval(noise_dict)
noise_dict['matched'] = 0


# stimfunction across two conditions for each subject 
stimfunc_all = []
for sid in range(8): # sid = 0
    # Create the stimulus time course of the conditions
    total_time = int(numTRs * trDuration)
    events = int(total_time / (event_duration + isi))
    onsets_A = []
    onsets_B = []
    randoized_label = np.repeat([1,2],int(events/2)).tolist()
    random.shuffle(randoized_label)
    for event_counter, cond in enumerate(randoized_label):
        
        # Flip a coin for each epoch to determine whether it is A or B
        if cond == 1:
            onsets_A.append(event_counter * (event_duration + isi))
        elif cond == 2:
            onsets_B.append(event_counter * (event_duration + isi))
            
    temporal_res = 1 # How many timepoints per second of the stim function are to be generated?
    
    # Create a time course of events 
    stimfunc_A = sim.generate_stimfunction(onsets=onsets_A,
                                           event_durations=[event_duration],
                                           total_time=total_time,
                                           temporal_resolution=temporal_res,
                                          )
    
    stimfunc_B = sim.generate_stimfunction(onsets=onsets_B,
                                           event_durations=[event_duration],
                                           total_time=total_time,
                                           temporal_resolution=temporal_res,
                                           )
    # stimfunc per subject
    stimfunc_ppt = np.concatenate((stimfunc_A, stimfunc_B), axis = 1)
    
    stimfunc_all.append(stimfunc_ppt)
    
    print('Load ROIs')
    nii_A = image.load_img(os.path.join(dat_dir, 'ROI_A_mpfc.nii.gz'))
    nii_B = image.load_img(os.path.join(dat_dir, 'ROI_B_precuenus.nii.gz'))
    ROI_A = nii_A.get_fdata()
    ROI_B = nii_B.get_fdata()
    
    # How many voxels per ROI
    voxels_A = int(ROI_A.sum())
    voxels_B = int(ROI_B.sum())
    
    # Create a pattern of activity across the two voxels
    print('Creating signal pattern')
  
    pattern_A = np.random.rand(voxels_A).reshape((voxels_A, 1))
    pattern_B = np.random.rand(voxels_B).reshape((voxels_B, 1))
   
    # Multiply each pattern by each voxel time course
    # Noise was added to the design matrix, to make the correlation pattern noise, so FCMA could be challenging. 
    # use normal distributed noise instead of unifrom distributed to make the data noisier 
    weights_A = np.tile(stimfunc_A, voxels_A) * pattern_A.T + np.random.normal(0,1.5, size = np.tile(stimfunc_A, voxels_A).shape) 
    weights_B = np.tile(stimfunc_B, voxels_B) * pattern_B.T + np.random.normal(0,1.5, size = np.tile(stimfunc_B, voxels_B).shape) 
        
    # Convolve the onsets with the HRF
    # TR less than feature is not good, but b/c this is simulated data, can ignore this concer. 
    print('Creating signal time course')
    signal_func_A = sim.convolve_hrf(stimfunction=weights_A,
                                   tr_duration=trDuration,
                                   temporal_resolution=temporal_res,
                                   scale_function=1,
                                   )
    
    signal_func_B = sim.convolve_hrf(stimfunction=weights_B,
                                   tr_duration=trDuration,
                                   temporal_resolution=temporal_res,
                                   scale_function=1,
                                   )
    
    
    # Multiply the signal by the signal change 
    signal_func_A =  signal_func_A * signal_change #+ signal_func_B * signal_change
    signal_func_B =  signal_func_B * signal_change #+ signal_func_A * signal_change

    # Combine the signal time course with the signal volume
    print('Creating signal volumes')
    signal_A = sim.apply_signal(signal_func_A,
                                ROI_A)
    
    signal_B = sim.apply_signal(signal_func_B,
                                ROI_B)
    
    # Combine the two signal timecourses
    signal = signal_A + signal_B
    
    # spare the true noise. 
    #print('Generating noise')
    #noise = sim.generate_noise(dimensions=dimensions,
    #                           stimfunction_tr=np.zeros((numTRs, 1)),
    #                           tr_duration=int(trDuration),
    #                           template=template,
    #                           mask=mask_cherry,
    #                           noise_dict=noise_dict,
    #                           temporal_proportion = 0.5)
    
    
    
    brain = signal #+ noise
    brain_nii = image.new_img_like(template_nii, brain, template_nii.affine)
    out_path = os.path.join(dat_dir, '../simulated_data')
    if os.path.isdir(out_path) is False:
        os.makedirs(out_path, exist_ok=True)
    brain_nii.to_filename(os.path.join(out_path, f"sub_{sid}_sim_dat.nii.gz"))
    
     
# write out the simulated epoch file 
sim.export_epoch_file(stimfunction = stimfunc_all,
                      filename = os.path.join(dat_dir, '../simulated_data/sim_epoch_file.npy'),
                      tr_duration = 1.0,
                      temporal_resolution = 1.0,)




#  Copyright 2016 Mingbo Cai, Princeton Neuroscience Instititute,
#  Princeton University
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


def test_can_instantiate():
    import brainiak.reprsimil.brsa
    s = brainiak.reprsimil.brsa.GBRSA()
    assert s, "Invalid GBRSA instance!"

    s = brainiak.reprsimil.brsa.GBRSA(n_iter=40, rank=4, auto_nuisance=False,
                                      nureg_method='PCA',
                                      baseline_single=False, logS_range=1.0,
                                      SNR_bins=11, rho_bins=40, tol=2e-3,
                                      optimizer='CG', random_state=0,
                                      anneal_speed=20, SNR_prior='unif')
    assert s, "Invalid GBRSA instance!"


def test_fit():
    from brainiak.reprsimil.brsa import GBRSA
    import brainiak.utils.utils as utils
    import scipy.stats
    import numpy as np
    import os.path
    np.random.seed(10)
    file_path = os.path.join(os.path.dirname(__file__), "example_design.1D")
    # Load an example design matrix
    design = utils.ReadDesign(fname=file_path)

    # concatenate it by 1, 2, and 3 times, mimicking different length
    # of experiments for different participants
    n_run = [2, 1, 1]
    design_mat = [None] * 3
    n_T = [None] * 3
    n_V = [40, 60, 60]
    for i in range(3):
        design_mat[i] = np.tile(design.design_task[:, :-1], [n_run[i], 1])

    # start simulating some data
    n_C = np.size(design_mat[0], axis=1)

    noise_bot = 0.5
    noise_top = 1.5
    noise_level = [None] * 3

    # AR(1) coefficient
    rho1_top = 0.8
    rho1_bot = -0.2
    rho1 = [None] * 3

    # generating noise
    noise = [None] * 3

    # baseline
    inten = [None] * 3
    for i in range(3):
        design_mat[i] = np.tile(design.design_task[:, :-1], [n_run[i], 1])
        n_T[i] = n_run[i] * design.n_TR
        noise_level[i] = np.random.rand(
            n_V[i]) * (noise_top - noise_bot) + noise_bot
        # noise level is random.
        rho1[i] = np.random.rand(n_V[i]) * (rho1_top - rho1_bot) + rho1_bot

        noise[i] = np.zeros([n_T[i], n_V[i]])
        noise[i][0, :] = np.random.randn(
            n_V[i]) * noise_level[i] / np.sqrt(1 - rho1[i]**2)
        for i_t in range(1, n_T[i]):
            noise[i][i_t, :] = noise[i][i_t - 1, :] * rho1[i] + \
                np.random.randn(n_V[i]) * noise_level[i]
        noise[i] = noise[i] + \
            np.dot(np.random.randn(n_T[i], 2), np.random.randn(2, n_V[i]))
        inten[i] = np.random.rand(n_V[i]) * 20.0

    # ideal covariance matrix
    ideal_cov = np.zeros([n_C, n_C])
    ideal_cov = np.eye(n_C) * 0.6
    ideal_cov[0:4, 0:4] = 0.2
    for cond in range(0, 4):
        ideal_cov[cond, cond] = 2
    ideal_cov[5:9, 5:9] = 0.9
    for cond in range(5, 9):
        ideal_cov[cond, cond] = 1
    L_full = np.linalg.cholesky(ideal_cov)

    # generating signal
    snr_top = 5.0  # test with high SNR
    snr_bot = 1.0
    # snr = np.random.rand(n_V)*(snr_top-snr_bot)+snr_bot
    # Notice that accurately speaking this is not snr. the magnitude of signal
    # depends not only on beta but also on x.

    snr = [None] * 3
    signal = [None] * 3
    betas_simulated = [None] * 3
    scan_onsets = [None] * 3
    Y = [None] * 3
    for i in range(3):
        snr[i] = np.random.rand(n_V[i]) * (snr_top - snr_bot) + snr_bot
        sqrt_v = noise_level[i] * snr[i]
        betas_simulated[i] = np.dot(
            L_full, np.random.randn(n_C, n_V[i])) * sqrt_v
        signal[i] = np.dot(design_mat[i], betas_simulated[i])

        # Adding noise to signal as data
        Y[i] = signal[i] + noise[i] + inten[i]

        scan_onsets[i] = np.linspace(0, n_T[i], num=n_run[i] + 1)

    # Test fitting.
    n_nureg = 2
    gbrsa = GBRSA(n_iter=15, auto_nuisance=True, logS_range=0.5, SNR_bins=11,
                  rho_bins=16, n_nureg=n_nureg, optimizer='L-BFGS-B')

    gbrsa.fit(X=Y, design=design_mat, scan_onsets=scan_onsets)

    # Check that result is significantly correlated with the ideal covariance
    # matrix
    u_b = gbrsa.U_
    u_i = ideal_cov
    p = scipy.stats.spearmanr(u_b[np.tril_indices_from(u_b)],
                              u_i[np.tril_indices_from(u_i)])[1]
    assert p < 0.01, (
        "Fitted covariance matrix does not correlate with ideal covariance "
        "matrix!")
    # check that the recovered SNRs makes sense
    p = scipy.stats.pearsonr(gbrsa.nSNR_[0], snr[0])[1]
    assert p < 0.01, "Fitted SNR does not correlate with simulated SNR!"
    p = scipy.stats.pearsonr(gbrsa.sigma_[1], noise_level[1])[1]
    assert p < 0.01, (
        "Fitted noise level does not correlate with simulated noise level!")
    p = scipy.stats.pearsonr(gbrsa.rho_[2], rho1[2])[1]
    assert p < 0.01, (
        "Fitted AR(1) coefficient does not correlate with simulated values!")
    assert np.shape(gbrsa.X0_[1]) == (n_T[1], n_nureg + 1), "Wrong size of X0"

    Y_new = [None] * 3
    noise_new = [None] * 3
    for i in range(3):
        noise_new[i] = np.zeros([n_T[i], n_V[i]])
        noise_new[i][0, :] = np.random.randn(
            n_V[i]) * noise_level[i] / np.sqrt(1 - rho1[i]**2)
        for i_t in range(1, n_T[i]):
            noise_new[i][i_t, :] = noise_new[i][i_t - 1, :] * \
                rho1[i] + np.random.randn(n_V[i]) * noise_level[i]

        Y_new[i] = signal[i] + noise_new[i] + inten[i]
    ts, ts0 = gbrsa.transform(Y_new, scan_onsets=scan_onsets)
    [score, score_null] = gbrsa.score(
        X=Y_new, design=design_mat, scan_onsets=scan_onsets)
    [score_noise, score_null_noise] = gbrsa.score(
        X=noise_new, design=design_mat, scan_onsets=scan_onsets)
    for i in range(3):
        assert np.shape(ts[i]) == (n_T[i], n_C) and np.shape(
            ts0[i]) == (n_T[i], n_nureg + 1)
        p = scipy.stats.pearsonr(ts[i][:, 0], design_mat[i][:, 0])[1]
        assert p < 0.01, (
            "Recovered time series does not correlate with true time series!")

        assert score[i] > score_null[i], (
            "Full model does not win over null model on data containing "
            "signal")

        assert score_noise[i] < score_null_noise[i], (
            "Null model does not win over full model on data without signal")

    [score, score_null] = gbrsa.score(
        X=[None] * 3, design=design_mat, scan_onsets=scan_onsets)
    assert score == [None] * 3 and score_null == [None] * \
        3, "score did not return list of None when data is None"
    ts, ts0 = gbrsa.transform(X=[None] * 3, scan_onsets=scan_onsets)
    assert ts == [None] * 3 and ts0 == [None] * \
        3, "transform did not return list of None when data is None"


def test_gradient():
    from brainiak.reprsimil.brsa import GBRSA
    import brainiak.utils.utils as utils
    import numpy as np
    import os.path
    import numdifftools as nd

    np.random.seed(100)
    file_path = os.path.join(os.path.dirname(__file__), "example_design.1D")
    # Load an example design matrix
    design = utils.ReadDesign(fname=file_path)

    # concatenate it by 1, 2, and 3 times, mimicking different length
    # of experiments for different participants
    n_run = [1, 2, 1]
    design_mat = [None] * 3
    n_T = [None] * 3
    n_V = [30, 30, 20]
    for i in range(3):
        design_mat[i] = np.tile(design.design_task[:, :-1], [n_run[i], 1])
        n_T[i] = n_run[i] * design.n_TR

    # start simulating some data
    n_C = np.size(design_mat[0], axis=1)

    noise_bot = 0.5
    noise_top = 1.5
    noise_level = [None] * 3
    for i in range(3):
        noise_level[i] = np.random.rand(
            n_V[i]) * (noise_top - noise_bot) + noise_bot
    # noise level is random.

    # AR(1) coefficient
    rho1_top = 0.8
    rho1_bot = -0.2
    rho1 = [None] * 3

    # generating noise
    noise = [None] * 3

    # baseline
    inten = [None] * 3
    for i in range(3):
        rho1[i] = np.random.rand(n_V[i]) * (rho1_top - rho1_bot) + rho1_bot
        noise[i] = np.zeros([n_T[i], n_V[i]])
        noise[i][0, :] = np.random.randn(
            n_V[i]) * noise_level[i] / np.sqrt(1 - rho1[i]**2)
        for i_t in range(1, n_T[i]):
            noise[i][i_t, :] = noise[i][i_t - 1, :] * rho1[i] + \
                np.random.randn(n_V[i]) * noise_level[i]
        noise[i] = noise[i] + \
            np.dot(np.random.randn(n_T[i], 2), np.random.randn(2, n_V[i]))
        inten[i] = np.random.rand(n_V[i]) * 20.0

    # ideal covariance matrix
    ideal_cov = np.zeros([n_C, n_C])
    ideal_cov = np.eye(n_C) * 0.6
    ideal_cov[0:4, 0:4] = 0.2
    for cond in range(0, 4):
        ideal_cov[cond, cond] = 2
    ideal_cov[5:9, 5:9] = 0.9
    for cond in range(5, 9):
        ideal_cov[cond, cond] = 1
    L_full = np.linalg.cholesky(ideal_cov)

    # generating signal
    snr_top = 5.0  # test with high SNR
    snr_bot = 1.0
    # snr = np.random.rand(n_V)*(snr_top-snr_bot)+snr_bot
    # Notice that accurately speaking this is not snr. the magnitude of signal
    # depends not only on beta but also on x.

    snr = [None] * 3
    signal = [None] * 3
    betas_simulated = [None] * 3
    scan_onsets = [None] * 3
    Y = [None] * 3
    for i in range(3):
        snr[i] = np.random.rand(n_V[i]) * (snr_top - snr_bot) + snr_bot
        sqrt_v = noise_level[i] * snr[i]
        betas_simulated[i] = np.dot(
            L_full, np.random.randn(n_C, n_V[i])) * sqrt_v
        signal[i] = np.dot(design_mat[i], betas_simulated[i])

        # Adding noise to signal as data
        Y[i] = signal[i] + noise[i] + inten[i]

        scan_onsets[i] = np.linspace(0, n_T[i], num=n_run[i] + 1)

    # Get some initial fitting.
    SNR_bins = 11
    rho_bins = 20
    gbrsa = GBRSA(n_iter=3, rank=n_C, SNR_bins=SNR_bins,
                  rho_bins=rho_bins, logS_range=0.5)

    n_grid = SNR_bins * rho_bins
    half_log_det_X0TAX0 = [np.random.randn(n_grid) for i in range(3)]
    log_weights = np.random.randn(n_grid)
    log_fixed_terms = [np.random.randn(n_grid) for i in range(3)]
    l_idx = np.tril_indices(n_C)
    L_vec = np.random.randn(int(n_C * (n_C + 1) / 2))
    n_X0 = [2, 2, 2]
    s = np.linspace(1, SNR_bins, n_grid)
    a = np.linspace(0.5, 1, n_grid)
    s2XTAcorrX = [None] * 3
    YTAcorrY_diag = [None] * 3
    sXTAcorrY = [None] * 3
    # The calculations below are quite arbitrary and do not conform
    # to the model. They simply conform to the symmetry property and shape of
    # the matrix indicated by the model
    for i in range(3):
        YTAcorrY_diag[i] = np.sum(Y[i] * Y[i], axis=0) * a[:, None]
        s2XTAcorrX[i] = np.dot(design_mat[i].T, design_mat[
                               i]) * s[:, None, None]**2 * a[:, None, None]
        sXTAcorrY[i] = np.dot(design_mat[i].T, Y[i]) * \
            s[:, None, None] * a[:, None, None]

    # test if the gradients are correct
    # print(log_fixed_terms)
    ll0, deriv0 = gbrsa._sum_loglike_marginalized(L_vec, s2XTAcorrX,
                                                  YTAcorrY_diag, sXTAcorrY,
                                                  half_log_det_X0TAX0,
                                                  log_weights, log_fixed_terms,
                                                  l_idx, n_C, n_T, n_V, n_X0,
                                                  n_grid, rank=None)
    # We test the gradient to the Cholesky factor
    vec = np.random.randn(np.size(L_vec))
    vec = vec / np.linalg.norm(vec)
    dd = nd.directionaldiff(
        lambda x: gbrsa._sum_loglike_marginalized(x, s2XTAcorrX, YTAcorrY_diag,
                                                  sXTAcorrY,
                                                  half_log_det_X0TAX0,
                                                  log_weights, log_fixed_terms,
                                                  l_idx, n_C, n_T, n_V, n_X0,
                                                  n_grid, rank=None)[0],
        L_vec,
        vec)
    assert np.isclose(dd, np.dot(deriv0, vec), rtol=1e-5), 'gradient incorrect'


def test_SNR_grids():
    import brainiak.reprsimil.brsa
    import numpy as np

    s = brainiak.reprsimil.brsa.GBRSA(SNR_prior='unif', SNR_bins=10)
    SNR_grids, SNR_weights = s._set_SNR_grids()
    assert (np.isclose(np.sum(SNR_weights), 1)
            and np.isclose(np.std(SNR_weights[1:-1]), 0)
            and np.all(SNR_weights > 0)
            and np.isclose(np.min(SNR_grids), 0)
            and np.all(SNR_grids >= 0)
            and np.isclose(np.max(SNR_grids), 1)
            ), 'SNR_weights or SNR_grids are incorrect for uniform prior'
    assert np.isclose(np.ptp(np.diff(SNR_grids[1:-1])), 0), \
        'SNR grids are not equally spaced for uniform prior'
    assert (np.size(SNR_grids) == np.size(SNR_weights)
            and np.size(SNR_grids) == 10
            ), ("size of SNR_grids or SNR_weights is not correct for uniform "
                "prior")

    s = brainiak.reprsimil.brsa.GBRSA(SNR_prior='lognorm', SNR_bins=35)
    SNR_grids, SNR_weights = s._set_SNR_grids()
    assert (np.all(SNR_grids >= 0)
            and np.isclose(np.sum(SNR_weights), 1)
            and np.all(SNR_weights > 0)
            and np.all(np.diff(SNR_grids) > 0)
            ), 'SNR_grids or SNR_weights not correct for log normal prior'

    s = brainiak.reprsimil.brsa.GBRSA(SNR_prior='exp')
    SNR_grids, SNR_weights = s._set_SNR_grids()
    assert (np.all(SNR_grids >= 0)
            and np.isclose(np.sum(SNR_weights), 1)
            and np.all(SNR_weights > 0)
            and np.all(np.diff(SNR_grids) > 0)
            ), 'SNR_grids or SNR_weights not correct for exponential prior'

    s = brainiak.reprsimil.brsa.GBRSA(SNR_prior='equal')
    SNR_grids, SNR_weights = s._set_SNR_grids()
    assert (np.all(SNR_grids == 1)
            and np.all(SNR_weights == 1)
            and np.size(SNR_grids) == 1
            ), 'SNR_grids or SNR_weights not correct for equal prior'


def test_n_nureg():
    import brainiak.reprsimil.brsa
    import numpy as np
    noise = np.dot(np.random.randn(100, 8), np.random.randn(
        8, 30)) + np.random.randn(100, 30) * 0.001
    design = np.random.randn(100, 2)
    s = brainiak.reprsimil.brsa.GBRSA(n_iter=2)
    s.fit(X=noise, design=design)
    assert s.n_nureg_[0] == 8, 'n_nureg_ estimation is wrong in GBRSA'


def test_grid_flatten_num_int():
    # Check for numeric integration of SNR, and correctly flattening 2-D grids
    # to 1-D grid.
    import brainiak.reprsimil.brsa
    import brainiak.utils.utils as utils
    import numpy as np
    import scipy.special
    n_V = 30
    n_T = 50
    n_C = 3
    design = np.random.randn(n_T, n_C)
    U_simu = np.asarray([[1.0, 0.1, 0.0], [0.1, 1.0, 0.2], [0.0, 0.2, 1.0]])
    L_simu = np.linalg.cholesky(U_simu)
    SNR = np.random.exponential(size=n_V)
    beta = np.dot(L_simu, np.random.randn(n_C, n_V)) * SNR
    noise = np.random.randn(n_T, n_V)
    Y = np.dot(design, beta) + noise
    X = design
    X_base = None
    scan_onsets = [0]

    s = brainiak.reprsimil.brsa.GBRSA(n_iter=1, auto_nuisance=False,
                                      SNR_prior='exp')
    s.fit(X=[Y], design=[design])
    rank = n_C
    l_idx, rank = s._chol_idx(n_C, rank)
    L = np.zeros((n_C, rank))
    n_l = np.size(l_idx[0])
    current_vec_U_chlsk_l = s.random_state_.randn(n_l) * 10
    L[l_idx] = current_vec_U_chlsk_l

    # Now we change the grids for SNR and rho for testing.
    s.SNR_bins = 2
    s.rho_bins = 2
    SNR_grids, SNR_weights = s._set_SNR_grids()
    # rho_grids, rho_weights = s._set_rho_grids()
    rho_grids = np.ones(2) * 0.1
    rho_weights = np.ones(2) / 2
    # We purposefully set all rhos to be equal to test flattening of
    # grids.
    n_grid = s.SNR_bins * s.rho_bins

    D, F, run_TRs, n_run = s._prepare_DF(
        n_T, scan_onsets=scan_onsets)
    XTY, XTDY, XTFY, YTY_diag, YTDY_diag, YTFY_diag, XTX, XTDX, XTFX \
        = s._prepare_data_XY(X, Y, D, F)
    X0TX0, X0TDX0, X0TFX0, XTX0, XTDX0, XTFX0, X0TY, X0TDY, X0TFY, X0, \
        X_base, n_X0, idx_DC = s._prepare_data_XYX0(
            X, Y, X_base, None, D, F, run_TRs, no_DC=False)

    X0TAX0, X0TAX0_i, XTAcorrX, XTAcorrY, YTAcorrY_diag, X0TAY, XTAX0 \
        = s._precompute_ar1_quad_forms_marginalized(
            XTY, XTDY, XTFY, YTY_diag, YTDY_diag, YTFY_diag, XTX,
            XTDX, XTFX, X0TX0, X0TDX0, X0TFX0, XTX0, XTDX0, XTFX0,
            X0TY, X0TDY, X0TFY, rho_grids, n_V, n_X0)

    half_log_det_X0TAX0, X0TAX0, X0TAX0_i, s2XTAcorrX, YTAcorrY_diag, \
        sXTAcorrY, X0TAY, XTAX0 = s._matrix_flattened_grid(
            X0TAX0, X0TAX0_i, SNR_grids, XTAcorrX, YTAcorrY_diag, XTAcorrY,
            X0TAY, XTAX0, n_C, n_V, n_X0, n_grid)

    assert (half_log_det_X0TAX0[0] == half_log_det_X0TAX0[1]
            and half_log_det_X0TAX0[2] == half_log_det_X0TAX0[3]
            and half_log_det_X0TAX0[0] == half_log_det_X0TAX0[2]
            ), '_matrix_flattened_grid has mistake with half_log_det_X0TAX0'
    assert (np.array_equal(X0TAX0[0, :, :], X0TAX0[1, :, :])
            and np.array_equal(X0TAX0[2, :, :], X0TAX0[3, :, :])
            and np.array_equal(X0TAX0[0, :, :], X0TAX0[2, :, :])
            ), '_matrix_flattened_grid has mistake X0TAX0'
    assert (np.array_equal(X0TAX0_i[0, :, :], X0TAX0_i[1, :, :])
            and np.array_equal(X0TAX0_i[2, :, :], X0TAX0_i[3, :, :])
            and np.array_equal(X0TAX0_i[0, :, :], X0TAX0_i[2, :, :])
            ), '_matrix_flattened_grid has mistake X0TAX0_i'
    assert np.allclose(
        np.dot(X0TAX0[0, :, :], X0TAX0_i[0, :, :]),
        np.eye(n_X0)
        ), 'X0TAX0_i is not inverse of X0TAX0'
    assert (np.array_equal(YTAcorrY_diag[0, :], YTAcorrY_diag[1, :])
            and np.array_equal(YTAcorrY_diag[2, :], YTAcorrY_diag[3, :])
            and np.array_equal(YTAcorrY_diag[0, :], YTAcorrY_diag[2, :])
            ), '_matrix_flattened_grid has mistake YTAcorrY_diag'
    assert (np.array_equal(sXTAcorrY[0, :, :], sXTAcorrY[1, :, :])
            and np.array_equal(sXTAcorrY[2, :, :], sXTAcorrY[3, :, :])
            and not np.array_equal(sXTAcorrY[0, :, :], sXTAcorrY[2, :, :])
            ), '_matrix_flattened_grid has mistake sXTAcorrY'
    assert (np.array_equal(X0TAY[0, :, :], X0TAY[1, :, :])
            and np.array_equal(X0TAY[2, :, :], X0TAY[3, :, :])
            and np.array_equal(X0TAY[0, :, :], X0TAY[2, :, :])
            ), '_matrix_flattened_grid has mistake X0TAY'
    assert (np.array_equal(XTAX0[0, :, :], XTAX0[1, :, :])
            and np.array_equal(XTAX0[2, :, :], XTAX0[3, :, :])
            and np.array_equal(XTAX0[0, :, :], XTAX0[2, :, :])
            ), '_matrix_flattened_grid has mistake XTAX0'

    # Now we test the other way
    rho_grids, rho_weights = s._set_rho_grids()
    # rho_grids, rho_weights = s._set_rho_grids()
    SNR_grids = np.ones(2) * 0.1
    SNR_weights = np.ones(2) / 2
    # We purposefully set all SNR to be equal to test flattening of
    # grids.
    X0TAX0, X0TAX0_i, XTAcorrX, XTAcorrY, YTAcorrY_diag, X0TAY, XTAX0 \
        = s._precompute_ar1_quad_forms_marginalized(
            XTY, XTDY, XTFY, YTY_diag, YTDY_diag, YTFY_diag, XTX,
            XTDX, XTFX, X0TX0, X0TDX0, X0TFX0, XTX0, XTDX0, XTFX0,
            X0TY, X0TDY, X0TFY, rho_grids, n_V, n_X0)

    half_log_det_X0TAX0, X0TAX0, X0TAX0_i, s2XTAcorrX, YTAcorrY_diag, \
        sXTAcorrY, X0TAY, XTAX0 = s._matrix_flattened_grid(
            X0TAX0, X0TAX0_i, SNR_grids, XTAcorrX, YTAcorrY_diag, XTAcorrY,
            X0TAY, XTAX0, n_C, n_V, n_X0, n_grid)

    assert (half_log_det_X0TAX0[0] == half_log_det_X0TAX0[2]
            and half_log_det_X0TAX0[1] == half_log_det_X0TAX0[3]
            and not half_log_det_X0TAX0[0] == half_log_det_X0TAX0[1]
            ), '_matrix_flattened_grid has mistake with half_log_det_X0TAX0'
    assert (np.array_equal(X0TAX0[0, :, :], X0TAX0[2, :, :])
            and np.array_equal(X0TAX0[1, :, :], X0TAX0[3, :, :])
            and not np.array_equal(X0TAX0[0, :, :], X0TAX0[1, :, :])
            ), '_matrix_flattened_grid has mistake X0TAX0'
    assert (np.array_equal(X0TAX0_i[0, :, :], X0TAX0_i[2, :, :])
            and np.array_equal(X0TAX0_i[1, :, :], X0TAX0_i[3, :, :])
            and not np.array_equal(X0TAX0_i[0, :, :], X0TAX0_i[1, :, :])
            ), '_matrix_flattened_grid has mistake X0TAX0_i'
    assert np.allclose(
        np.dot(X0TAX0[0, :, :], X0TAX0_i[0, :, :]),
        np.eye(n_X0)
        ), 'X0TAX0_i is not inverse of X0TAX0'
    assert (np.array_equal(YTAcorrY_diag[0, :], YTAcorrY_diag[2, :])
            and np.array_equal(YTAcorrY_diag[1, :], YTAcorrY_diag[3, :])
            and not np.array_equal(YTAcorrY_diag[0, :],
                                   YTAcorrY_diag[1, :])
            ), '_matrix_flattened_grid has mistake YTAcorrY_diag'
    assert (np.array_equal(sXTAcorrY[0, :, :], sXTAcorrY[2, :, :])
            and np.array_equal(sXTAcorrY[1, :, :], sXTAcorrY[3, :, :])
            and not np.array_equal(sXTAcorrY[0, :, :], sXTAcorrY[1, :, :])
            ), '_matrix_flattened_grid has mistake sXTAcorrY'
    assert (np.array_equal(X0TAY[0, :, :], X0TAY[2, :, :])
            and np.array_equal(X0TAY[1, :, :], X0TAY[3, :, :])
            and not np.array_equal(X0TAY[0, :, :], X0TAY[1, :, :])
            ), '_matrix_flattened_grid has mistake X0TAY'
    assert (np.array_equal(XTAX0[0, :, :], XTAX0[2, :, :])
            and np.array_equal(XTAX0[1, :, :], XTAX0[3, :, :])
            and not np.array_equal(XTAX0[0, :, :], XTAX0[1, :, :])
            ), '_matrix_flattened_grid has mistake XTAX0'

    # Now test the integration over SNR
    s.SNR_bins = 50
    s.rho_bins = 1
    SNR_grids, SNR_weights = s._set_SNR_grids()
    rho_grids, rho_weights = s._set_rho_grids()
    n_grid = s.SNR_bins * s.rho_bins

    def setup_for_test():
        # This function will be re-used to set up the variables necessary for
        # testing.

        X0TAX0, X0TAX0_i, XTAcorrX, XTAcorrY, YTAcorrY_diag, X0TAY, XTAX0 \
            = s._precompute_ar1_quad_forms_marginalized(
                XTY, XTDY, XTFY, YTY_diag, YTDY_diag, YTFY_diag, XTX,
                XTDX, XTFX, X0TX0, X0TDX0, X0TFX0, XTX0, XTDX0, XTFX0,
                X0TY, X0TDY, X0TFY, rho_grids, n_V, n_X0)

        half_log_det_X0TAX0, X0TAX0, X0TAX0_i, s2XTAcorrX, YTAcorrY_diag, \
            sXTAcorrY, X0TAY, XTAX0 = s._matrix_flattened_grid(
                X0TAX0, X0TAX0_i, SNR_grids, XTAcorrX, YTAcorrY_diag, XTAcorrY,
                X0TAY, XTAX0, n_C, n_V, n_X0, n_grid)

        log_weights = np.reshape(
            np.log(SNR_weights[:, None]) + np.log(rho_weights), n_grid)
        all_rho_grids = np.reshape(np.repeat(
            rho_grids[None, :], s.SNR_bins, axis=0), n_grid)
        log_fixed_terms = - (n_T - n_X0) / 2 * np.log(2 * np.pi) + n_run \
            / 2 * np.log(1 - all_rho_grids**2) + scipy.special.gammaln(
                (n_T - n_X0 - 2) / 2) + (n_T - n_X0 - 2) / 2 * np.log(2)
        return s2XTAcorrX, YTAcorrY_diag, sXTAcorrY, half_log_det_X0TAX0, \
            log_weights, log_fixed_terms

    (s2XTAcorrX, YTAcorrY_diag, sXTAcorrY, half_log_det_X0TAX0, log_weights,
     log_fixed_terms) = setup_for_test()
    LL_total, _ = s._loglike_marginalized(current_vec_U_chlsk_l, s2XTAcorrX,
                                          YTAcorrY_diag, sXTAcorrY,
                                          half_log_det_X0TAX0, log_weights,
                                          log_fixed_terms, l_idx, n_C, n_T,
                                          n_V, n_X0, n_grid, rank=rank)
    LL_total = - LL_total
    # Now we re-calculate using scipy.integrate
    s.SNR_bins = 100
    SNR_grids = np.linspace(0, 12, s.SNR_bins)
    SNR_weights = np.exp(- SNR_grids)
    SNR_weights = SNR_weights / np.sum(SNR_weights)
    n_grid = s.SNR_bins * s.rho_bins

    (s2XTAcorrX, YTAcorrY_diag, sXTAcorrY, half_log_det_X0TAX0, log_weights,
     log_fixed_terms) = setup_for_test()
    LL_raw, _, _, _ = s._raw_loglike_grids(L, s2XTAcorrX, YTAcorrY_diag,
                                           sXTAcorrY, half_log_det_X0TAX0,
                                           log_weights, log_fixed_terms,
                                           n_C, n_T, n_V, n_X0,
                                           n_grid, rank)
    result_sum, max_value, result_exp = utils.sumexp_stable(LL_raw)
    scipy_sum = scipy.integrate.simpson(y=result_exp, axis=0)
    LL_total_scipy = np.sum(np.log(scipy_sum) + max_value)

    tol = 1e-3
    assert (np.isclose(LL_total_scipy, LL_total, rtol=tol)), \
        'Error of log likelihood calculation exceeds the tolerance'

    # Now test the log normal prior
    s = brainiak.reprsimil.brsa.GBRSA(n_iter=1, auto_nuisance=False,
                                      SNR_prior='lognorm')
    s.SNR_bins = 50
    s.rho_bins = 1
    SNR_grids, SNR_weights = s._set_SNR_grids()
    rho_grids, rho_weights = s._set_rho_grids()
    n_grid = s.SNR_bins * s.rho_bins

    (s2XTAcorrX, YTAcorrY_diag, sXTAcorrY, half_log_det_X0TAX0, log_weights,
     log_fixed_terms) = setup_for_test()
    LL_total, _ = s._loglike_marginalized(current_vec_U_chlsk_l, s2XTAcorrX,
                                          YTAcorrY_diag, sXTAcorrY,
                                          half_log_det_X0TAX0, log_weights,
                                          log_fixed_terms, l_idx, n_C, n_T,
                                          n_V, n_X0, n_grid, rank=rank)
    LL_total = - LL_total
    # Now we re-calculate using scipy.integrate
    s.SNR_bins = 400
    SNR_grids = np.linspace(1e-8, 20, s.SNR_bins)
    log_SNR_weights = scipy.stats.lognorm.logpdf(SNR_grids, s=s.logS_range)
    result_sum, max_value, result_exp = utils.sumexp_stable(
        log_SNR_weights[:, None])
    SNR_weights = np.squeeze(result_exp / result_sum)
    n_grid = s.SNR_bins * s.rho_bins

    (s2XTAcorrX, YTAcorrY_diag, sXTAcorrY, half_log_det_X0TAX0, log_weights,
     log_fixed_terms) = setup_for_test()
    LL_raw, _, _, _ = s._raw_loglike_grids(L, s2XTAcorrX, YTAcorrY_diag,
                                           sXTAcorrY, half_log_det_X0TAX0,
                                           log_weights, log_fixed_terms,
                                           n_C, n_T, n_V, n_X0,
                                           n_grid, rank)
    result_sum, max_value, result_exp = utils.sumexp_stable(LL_raw)
    scipy_sum = scipy.integrate.simpson(y=result_exp, axis=0)
    LL_total_scipy = np.sum(np.log(scipy_sum) + max_value)

    tol = 1e-3
    assert (np.isclose(LL_total_scipy, LL_total, rtol=tol)), \
        'Error of log likelihood calculation exceeds the tolerance'



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Hierarchical Topographical Factor Analysis (HTFA)

This implementation is based on the work in [Manning2014-1]_, [Manning2014-2]_,
[AndersonMJ2016]_, and [Manning2018]_.

.. [Manning2014-1] "Topographic factor analysis: a bayesian model for
   inferring brain networks from neural data", J. R. Manning,
   R. Ranganath, K. A. Norman, and D. M. Blei. PLoS One, vol. 9, no. 5,
   2014.

.. [Manning2014-2] "Hierarchical topographic factor analysis", Jeremy. R.
   Manning, R. Ranganath, W. Keung, N. B. Turk-Browne, J. D.Cohen,
   K. A. Norman, and D. M. Blei. Pattern Recognition in Neuroimaging,
   2014 International Workshop on, June 2014.

.. [Manning2018] "A Probabilistic Approach to Discovering Dynamic Full-brain
   Functional Connectivit Patterns", J. R. Manning, X. Zhu, T.L. Willke,
   R. Ranganath, K. Stachenfeld, U. Hasson, D. M. Blei and K. A. Norman.
   Neuroimage, 2018.
   https://doi.org/10.1016/j.neuroimage.2018.01.071

.. [AndersonMJ2016] "Enabling Factor Analysis on Thousand-Subject Neuroimaging
   Datasets",
   Michael J. Anderson, Mihai Capotă, Javier S. Turek, Xia Zhu, Theodore L.
   Willke, Yida Wang, Po-Hsuan Chen, Jeremy R. Manning, Peter J. Ramadge,
   Kenneth A. Norman,
   IEEE International Conference on Big Data, 2016.
   https://doi.org/10.1109/BigData.2016.7840719
"""

# Authors: Xia Zhu (Intel Labs), Jeremy Manning (Dartmouth College) 2015~2016

import numpy as np
from mpi4py import MPI
from scipy.optimize import linear_sum_assignment
from sklearn.metrics import mean_squared_error
from scipy.spatial import distance
import logging
from .tfa import TFA
from ..utils.utils import from_tri_2_sym, from_sym_2_tri

__all__ = [
    "HTFA",
]

logger = logging.getLogger(__name__)


class HTFA(TFA):
    """Hierarchical Topographical Factor Analysis (HTFA)

    Given multi-subject data, HTFA factorizes data from each subject as a
    spatial factor F and a weight matrix W per subject. Also at top
    level, it estimates global template across subjects:


    Parameters
    ----------

    K : int
        Number of factors to compute.

    n_subj : int
        Total number of subjects in dataset.

    max_global_iter : int, default: 10
        Number of global iterations to run the algorithm.

    max_local_iter : int, default: 10
        Number of local iterations to run on each subject within each
        global interation.

    threshold : float, default: 1.0
        Tolerance for terminate the parameter estimation

    nlss_method : {'trf', 'dogbox', 'lm'}, default: 'trf'
        Non-Linear Least Square (NLSS) algorithm used by scipy.least_suqares to
        perform minimization. More information at
        http://docs.scipy.org/doc/scipy-0.17.0/reference/generated/scipy.optimize.least_squares.html

    nlss_loss: str or callable, default: 'linear'
        Loss function used by scipy.least_squares.
        More information at
        http://docs.scipy.org/doc/scipy-0.17.0/reference/generated/scipy.optimize.least_squares.html

    jac : {'2-point', '3-point', 'cs', callable}, default: '2-point'
        Method of computing the Jacobian matrix.
        More information at
        http://docs.scipy.org/doc/scipy-0.17.0/reference/generated/scipy.optimize.least_squares.html

    x_scale : float or array_like or 'jac', default: 1.0
        Characteristic scale of each variable for scipy.least_suqares.
        More information at
        http://docs.scipy.org/doc/scipy-0.17.0/reference/generated/scipy.optimize.least_squares.html

    tr_solver: {None, 'exact', 'lsmr'}, default: None
        Method for solving trust-region subproblems, relevant only for 'trf'
        and 'dogbox' methods.
        More information at
        http://docs.scipy.org/doc/scipy-0.17.0/reference/generated/scipy.optimize.least_squares.html

    weight_method : {'rr','ols'}, default: 'rr'
        Method for estimating weight matrix W given X and F.
        'rr' means ridge regression, 'ols' means ordinary least square.

    upper_ratio : float, default: 1.8
        The upper bound of the ratio between factor's width and brain diameter.

    lower_ratio : float, default: 0.02
        The lower bound of the ratio between factor's width and brain diameter.

    voxel_ratio : float, default: 0.25
        The percentage of voxels to sample in each inner iteration.

    tr_ratio : float, default: 0.1
        The percentage of trs to sample in each inner iteration.

    max_voxel : int, default: 5000
        The maximum number of voxels to sample in each inner iteration.

    max_tr : int, default: 500
        The maximum number of trs to sample in each inner iteration.

    comm : Intracomm
        MPI communication group, default MPI.COMM_WORLD

    verbose : boolean, default: False
        Verbose mode flag.


    Attributes
    ----------
    global_prior_ : 1D array
        The global prior on mean and variance of centers and widths.

    global_posterior_ : 1D array
        The global posterior on mean and variance of centers and widths.

    local_posterior_ : 1D array
        Local posterior on centers and widths of subjects allocated
        to this process.

    local_weights_ : 1D array
        Local posterior on weights allocated to this process.


    Notes
    -----
    We recommend to use data in MNI space to better interpret global template

    """

    def __init__(self, K, n_subj, max_global_iter=10, max_local_iter=10,
                 threshold=0.01, nlss_method='trf', nlss_loss='soft_l1',
                 jac='2-point', x_scale='jac', tr_solver=None,
                 weight_method='rr', upper_ratio=1.8, lower_ratio=0.02,
                 voxel_ratio=0.25, tr_ratio=0.1, max_voxel=5000, max_tr=500,
                 comm=MPI.COMM_WORLD, verbose=False):
        self.K = K
        self.n_subj = n_subj
        self.max_global_iter = max_global_iter
        self.max_local_iter = max_local_iter
        self.threshold = threshold
        self.nlss_method = nlss_method
        self.nlss_loss = nlss_loss
        self.jac = jac
        self.x_scale = x_scale
        self.tr_solver = tr_solver
        self.weight_method = weight_method
        self.upper_ratio = upper_ratio
        self.lower_ratio = lower_ratio
        self.voxel_ratio = voxel_ratio
        self.tr_ratio = tr_ratio
        self.max_voxel = max_voxel
        self.max_tr = max_tr
        self.comm = comm
        self.verbose = verbose

    def _converged(self):
        """Check convergence based on maximum absolute difference

        Returns
        -------

        converged : boolean
            Whether the parameter estimation converged.

        max_diff : float
            Maximum absolute difference between prior and posterior.

        """

        prior = self.global_prior_[0:self.prior_size]
        posterior = self.global_posterior_[0:self.prior_size]
        diff = prior - posterior
        max_diff = np.max(np.fabs(diff))
        if self.verbose:
            _, mse = self._mse_converged()
            diff_ratio = np.sum(diff ** 2) / np.sum(posterior ** 2)
            logger.info(
                'htfa prior posterior max diff %f mse %f diff_ratio %f' %
                ((max_diff, mse, diff_ratio)))

        if max_diff > self.threshold:
            return False, max_diff
        else:
            return True, max_diff

    def _mse_converged(self):
        """Check convergence based on mean squared difference between
            prior and posterior

        Returns
        -------

        converged : boolean
            Whether the parameter estimation converged.

        mse : float
            Mean squared error between prior and posterior.

        """

        prior = self.global_prior_[0:self.prior_size]
        posterior = self.global_posterior_[0:self.prior_size]
        mse = mean_squared_error(prior, posterior,
                                 multioutput='uniform_average')
        if mse > self.threshold:
            return False, mse
        else:
            return True, mse

    def _map_update(
            self,
            prior_mean,
            prior_cov,
            global_cov_scaled,
            new_observation):
        """Maximum A Posterior (MAP) update of a parameter

        Parameters
        ----------

        prior_mean : float or 1D array
            Prior mean of parameters.

        prior_cov : float or 1D array
            Prior variance of scalar parameter, or
            prior covariance of multivariate parameter

        global_cov_scaled : float or 1D array
            Global prior variance of scalar parameter, or
            global prior covariance of multivariate parameter

        new_observation : 1D or 2D array, with shape [n_dim, n_subj]
            New observations on parameters.

        Returns
        -------

        posterior_mean : float or 1D array
            Posterior mean of parameters.

        posterior_cov : float or 1D array
            Posterior variance of scalar parameter, or
            posterior covariance of multivariate parameter

        """
        common = np.linalg.inv(prior_cov + global_cov_scaled)
        observation_mean = np.mean(new_observation, axis=1)
        posterior_mean = prior_cov.dot(common.dot(observation_mean)) +\
            global_cov_scaled.dot(common.dot(prior_mean))
        posterior_cov =\
            prior_cov.dot(common.dot(global_cov_scaled))
        return posterior_mean, posterior_cov

    def _map_update_posterior(self):
        """Maximum A Posterior (MAP) update of HTFA parameters

        Returns
        -------
        HTFA
            Returns the instance itself.
        """
        self.global_posterior_ = self.global_prior_.copy()
        prior_centers = self.get_centers(self.global_prior_)
        prior_widths = self.get_widths(self.global_prior_)
        prior_centers_mean_cov = self.get_centers_mean_cov(self.global_prior_)
        prior_widths_mean_var = self.get_widths_mean_var(self.global_prior_)
        center_size = self.K * self.n_dim
        posterior_size = center_size + self.K
        for k in np.arange(self.K):
            next_centers = np.zeros((self.n_dim, self.n_subj))
            next_widths = np.zeros(self.n_subj)
            for s in np.arange(self.n_subj):
                center_start = s * posterior_size
                width_start = center_start + center_size
                start_idx = center_start + k * self.n_dim
                end_idx = center_start + (k + 1) * self.n_dim
                next_centers[:, s] = self.gather_posterior[start_idx:end_idx]\
                    .copy()
                next_widths[s] = self.gather_posterior[width_start + k].copy()

            # centers
            posterior_mean, posterior_cov = self._map_update(
                prior_centers[k].T.copy(),
                from_tri_2_sym(prior_centers_mean_cov[k], self.n_dim),
                self.global_centers_cov_scaled,
                next_centers)
            self.global_posterior_[k * self.n_dim:(k + 1) * self.n_dim] =\
                posterior_mean.T
            start_idx = self.map_offset[2] + k * self.cov_vec_size
            end_idx = self.map_offset[2] + (k + 1) * self.cov_vec_size
            self.global_posterior_[start_idx:end_idx] =\
                from_sym_2_tri(posterior_cov)

            # widths
            common = 1.0 /\
                (prior_widths_mean_var[k] + self.global_widths_var_scaled)
            observation_mean = np.mean(next_widths)
            tmp = common * self.global_widths_var_scaled
            self.global_posterior_[self.map_offset[1].item() + k] = \
                prior_widths_mean_var[k] * common * observation_mean +\
                tmp * prior_widths[k]
            self.global_posterior_[self.map_offset[3].item() + k] = \
                prior_widths_mean_var[k] * tmp

        return self

    def _get_gather_offset(self, size):
        """Calculate the offset for gather result from this process

        Parameters
        ----------

        size : int
            The total number of process.

        Returns
        -------

        tuple_size : tuple_int
            Number of elements to send from each process
            (one integer for each process)

        tuple_offset : tuple_int
            Number of elements away from the first element
            in the array at which to begin the new, segmented
            array for a process
            (one integer for each process)

        subject_map : dictionary
            Mapping between global subject id to local id

        """

        gather_size = np.zeros(size).astype(int)
        gather_offset = np.zeros(size).astype(int)
        num_local_subjs = np.zeros(size).astype(int)
        subject_map = {}

        for idx, s in enumerate(np.arange(self.n_subj)):
            cur_rank = idx % size
            gather_size[cur_rank] += self.prior_size
            subject_map[idx] = (cur_rank, num_local_subjs[cur_rank])
            num_local_subjs[cur_rank] += 1

        for idx in np.arange(size - 1) + 1:
            gather_offset[idx] = gather_offset[idx - 1] + gather_size[idx - 1]

        tuple_size = tuple(gather_size)
        tuple_offset = tuple(gather_offset)
        return tuple_size, tuple_offset, subject_map

    def _get_weight_size(self, data, n_local_subj):
        """Calculate the size of weight for this process

        Parameters
        ----------

        data : a list of 2D array, each in shape [n_voxel, n_tr]
            The fMRI data from multi-subject.

        n_local_subj : int
            Number of subjects allocated to this process.


        Returns
        -------

        weight_size : 1D array
            The size of total subject weight on this process.

        local_weight_offset : 1D array
            Number of elements away from the first element
            in the combined weight array at which to begin
            the new, segmented array for a subject

        """

        weight_size = np.zeros(1).astype(int)
        local_weight_offset = np.zeros(n_local_subj).astype(int)
        for idx, subj_data in enumerate(data):
            if idx > 0:
                local_weight_offset[idx] = weight_size[0]
            weight_size[0] += self.K * subj_data.shape[1]
        return weight_size, local_weight_offset

    def _get_subject_info(self, n_local_subj, data):
        """Calculate metadata for subjects allocated to this process

        Parameters
        ----------

        n_local_subj : int
            Number of subjects allocated to this process.

        data : list of 2D array. Each in shape [n_voxel, n_tr]
            Total number of MPI process.

        Returns
        -------

        max_sample_tr : 1D array
            Maximum number of TR to subsample for each subject

        max_sample_voxel : 1D array
            Maximum number of voxel to subsample for each subject

        """

        max_sample_tr = np.zeros(n_local_subj).astype(int)
        max_sample_voxel = np.zeros(n_local_subj).astype(int)
        for idx in np.arange(n_local_subj):
            nvoxel = data[idx].shape[0]
            ntr = data[idx].shape[1]
            max_sample_voxel[idx] =\
                min(self.max_voxel, int(self.voxel_ratio * nvoxel))
            max_sample_tr[idx] = min(self.max_tr, int(self.tr_ratio * ntr))
        return max_sample_tr, max_sample_voxel

    def _get_mpi_info(self):
        """get basic MPI info

        Returns
        -------
        comm : Intracomm
            Returns MPI communication group

        rank : integer
            Returns the rank of this process

        size : integer
            Returns total number of processes

        """

        rank = self.comm.Get_rank()
        size = self.comm.Get_size()
        return rank, size

    def _init_prior_posterior(self, rank, R, n_local_subj):
        """set prior for this subject

        Parameters
        ----------

        rank : integer
            The rank of this process

        R : list of 2D arrays, element i has shape=[n_voxel, n_dim]
            Each element in the list contains the scanner coordinate matrix
            of fMRI data of one subject.

        n_local_subj : integer
            The number of subjects allocated to this process.


        Returns
        -------
        HTFA
            Returns the instance itself.

        """

        if rank == 0:
            idx = np.random.choice(n_local_subj, 1)
            self.global_prior_, self.global_centers_cov, \
                self.global_widths_var = self.get_template(R[idx[0]])
            self.global_centers_cov_scaled =\
                self.global_centers_cov / float(self.n_subj)
            self.global_widths_var_scaled =\
                self.global_widths_var / float(self.n_subj)
            self.gather_posterior = np.zeros(self.n_subj * self.prior_size)
            self.global_posterior_ = np.zeros(self.prior_size)
        else:
            self.global_prior_ = np.zeros(self.prior_bcast_size)
            self.global_posterior_ = None
            self.gather_posterior = None
        return self

    def _gather_local_posterior(self, use_gather,
                                gather_size, gather_offset):
        """Gather/Gatherv local posterior


        Parameters
        ----------
        comm : object
            MPI communication group

        use_gather : boolean
            Whether to use Gather or Gatherv

        gather_size : 1D array
            The size of each local posterior

        gather_offset : 1D array
            The offset of each local posterior


        Returns
        -------
        HTFA
            Returns the instance itself.


        Notes
        -----
        We use numpy array rather than generic Python objects for MPI
        communication because Gatherv is only supported for the former.
        https://pythonhosted.org/mpi4py/usrman/tutorial.html

        """
        if use_gather:
            self.comm.Gather(self.local_posterior_,
                             self.gather_posterior, root=0)
        else:
            target = [
                self.gather_posterior,
                gather_size,
                gather_offset,
                MPI.DOUBLE]
            self.comm.Gatherv(self.local_posterior_, target)
        return self

    def _assign_posterior(self):
        """assign posterior to the right prior based on
           Hungarian algorithm

        Returns
        -------
        HTFA
            Returns the instance itself.
        """

        prior_centers = self.get_centers(self.global_prior_)
        posterior_centers = self.get_centers(self.global_posterior_)
        posterior_widths = self.get_widths(self.global_posterior_)
        posterior_centers_mean_cov =\
            self.get_centers_mean_cov(self.global_posterior_)
        posterior_widths_mean_var =\
            self.get_widths_mean_var(self.global_posterior_)
        # linear assignment on centers
        cost = distance.cdist(prior_centers, posterior_centers, 'euclidean')
        _, col_ind = linear_sum_assignment(cost)
        # reorder centers/widths based on cost assignment
        self.set_centers(self.global_posterior_, posterior_centers)
        self.set_widths(self.global_posterior_, posterior_widths)
        # reorder cov/var based on cost assignment
        self.set_centers_mean_cov(
            self.global_posterior_,
            posterior_centers_mean_cov[col_ind])
        self.set_widths_mean_var(
            self.global_posterior_,
            posterior_widths_mean_var[col_ind])
        return self

    def _update_global_posterior(
            self, rank, m, outer_converged):
        """Update global posterior and then check convergence

        Parameters
        ----------

        rank : integer
            The rank of current process.

        m : integer
            The outer iteration number of HTFA.

        outer_converged : 1D array
            Record whether HTFA loop converged


        Returns
        -------
        1D array, contains only 1 element for MPI
            1 means HTFA converged, 0 means not converged.

        """
        if rank == 0:
            self._map_update_posterior()
            self._assign_posterior()
            is_converged, _ = self._converged()
            if is_converged:
                logger.info("converged at %d outer iter" % (m))
                outer_converged[0] = 1
            else:
                self.global_prior_ = self.global_posterior_
        return outer_converged

    def _update_weight(self, data, R, n_local_subj, local_weight_offset):
        """update local weight

        Parameters
        ----------

        data : list of 2D array, element i has shape=[n_voxel, n_tr]
            Subjects' fMRI data.

        R : list of 2D arrays, element i has shape=[n_voxel, n_dim]
            Each element in the list contains the scanner coordinate matrix
            of fMRI data of one subject.

        n_local_subj : integer
            Number of subjects allocated to this process.

        local_weight_offset : 1D array
            Offset of each subject's weights on this process.


        Returns
        -------
        HTFA
            Returns the instance itself.

        """
        for s, subj_data in enumerate(data):
            base = s * self.prior_size
            centers = self.local_posterior_[base:base + self.K * self.n_dim]\
                .reshape((self.K, self.n_dim))
            start_idx = base + self.K * self.n_dim
            end_idx = base + self.prior_size
            widths = self.local_posterior_[start_idx:end_idx]\
                .reshape((self.K, 1))
            unique_R, inds = self.get_unique_R(R[s])
            F = self.get_factors(unique_R, inds, centers, widths)
            start_idx = local_weight_offset[s]
            if s == n_local_subj - 1:
                self.local_weights_[start_idx:] =\
                    self.get_weights(subj_data, F).ravel()
            else:
                end_idx = local_weight_offset[s + 1]
                self.local_weights_[start_idx:end_idx] =\
                    self.get_weights(subj_data, F).ravel()
        return self

    def _fit_htfa(self, data, R):
        """HTFA main algorithm

        Parameters
        ----------

        data : list of 2D array. Each in shape [n_voxel, n_tr]
            The fMRI data from multiple subjects.

        R : list of 2D arrays, element i has shape=[n_voxel, n_dim]
            Each element in the list contains the scanner coordinate matrix
            of fMRI data of one subject.

        Returns
        -------
        HTFA
            Returns the instance itself.
        """

        rank, size = self._get_mpi_info()
        use_gather = True if self.n_subj % size == 0 else False
        n_local_subj = len(R)
        max_sample_tr, max_sample_voxel =\
            self._get_subject_info(n_local_subj, data)

        tfa = []
        # init tfa for each subject
        for s, subj_data in enumerate(data):
            tfa.append(TFA(
                max_iter=self.max_local_iter,
                threshold=self.threshold,
                K=self.K,
                nlss_method=self.nlss_method,
                nlss_loss=self.nlss_loss,
                x_scale=self.x_scale,
                tr_solver=self.tr_solver,
                weight_method=self.weight_method,
                upper_ratio=self.upper_ratio,
                lower_ratio=self.lower_ratio,
                verbose=self.verbose,
                max_num_tr=max_sample_tr[s],
                max_num_voxel=max_sample_voxel[s]))

        # map data to processes
        gather_size, gather_offset, subject_map =\
            self._get_gather_offset(size)
        self.local_posterior_ = np.zeros(n_local_subj * self.prior_size)
        self._init_prior_posterior(rank, R, n_local_subj)
        node_weight_size, local_weight_offset =\
            self._get_weight_size(data, n_local_subj)
        self.local_weights_ = np.zeros(node_weight_size[0])

        m = 0
        outer_converged = np.array([0])
        while m < self.max_global_iter and not outer_converged[0]:
            if (self.verbose):
                logger.info("HTFA global iter %d " % (m))
            # root broadcast first 4 fields of global_prior to all nodes
            self.comm.Bcast(self.global_prior_, root=0)
            # each node loop over its data
            for s, subj_data in enumerate(data):
                # update tfa with current local prior
                tfa[s].set_prior(self.global_prior_[0:self.prior_size].copy())
                tfa[s].set_seed(m * self.max_local_iter)
                tfa[s].fit(
                    subj_data,
                    R=R[s],
                    template_prior=self.global_prior_.copy())
                tfa[s]._assign_posterior()
                start_idx = s * self.prior_size
                end_idx = (s + 1) * self.prior_size
                self.local_posterior_[start_idx:end_idx] =\
                    tfa[s].local_posterior_

            self._gather_local_posterior(
                use_gather,
                gather_size,
                gather_offset)

            # root updates global_posterior
            outer_converged =\
                self._update_global_posterior(rank, m, outer_converged)
            self.comm.Bcast(outer_converged, root=0)
            m += 1

        # update weight matrix for each subject
        self._update_weight(
            data,
            R,
            n_local_subj,
            local_weight_offset)

        return self

    def _check_input(self, X, R):
        """Check whether input data and coordinates in right type

        Parameters
        ----------
        X :  list of 2D arrays, element i has shape=[voxels_i, samples]
            Each element in the list contains the fMRI data of one subject.

        R : list of 2D arrays, element i has shape=[n_voxel, n_dim]
            Each element in the list contains the scanner coordinate matrix
            of fMRI data of one subject.

        Returns
        -------
        HTFA
            Returns the instance itself.
        """
        # Check data type
        if not isinstance(X, list):
            raise TypeError("Input data should be a list")

        if not isinstance(R, list):
            raise TypeError("Coordinates should be a list")

        # Check the number of subjects
        if len(X) < 1:
            raise ValueError("Need at leat one subject to train the model.\
                              Got {0:d}".format(len(X)))

        for idx, x in enumerate(X):
            if not isinstance(x, np.ndarray):
                raise TypeError("Each subject data should be an array")
            if x.ndim != 2:
                raise TypeError("Each subject data should be 2D array")
            if not isinstance(R[idx], np.ndarray):
                raise TypeError(
                    "Each scanner coordinate matrix should be an array")
            if R[idx].ndim != 2:
                raise TypeError(
                    "Each scanner coordinate matrix should be 2D array")
            if x.shape[0] != R[idx].shape[0]:
                raise TypeError(
                    "n_voxel should be the same in X[idx] and R[idx]")
        return self

    def fit(self, X, R):
        """Compute Hierarchical Topographical Factor Analysis Model
           [Manning2014-1][Manning2014-2]

        Parameters
        ----------
        X :  list of 2D arrays, element i has shape=[voxels_i, samples]
            Each element in the list contains the fMRI data of one subject.

        R : list of 2D arrays, element i has shape=[n_voxel, n_dim]
            Each element in the list contains the scanner coordinate matrix
            of fMRI data of one subject.

        Returns
        -------
        HTFA
            Returns the instance itself.
        """
        self._check_input(X, R)
        if self.verbose:
            logger.info("Start to fit HTFA")
        self.n_dim = R[0].shape[1]
        self.cov_vec_size = np.sum(np.arange(self.n_dim) + 1)
        # centers,widths
        self.prior_size = self.K * (self.n_dim + 1)
        # centers,widths,centerCov,widthVar
        self.prior_bcast_size =\
            self.K * (self.n_dim + 2 + self.cov_vec_size)
        self.get_map_offset()
        self._fit_htfa(X, R)
        return self



"""Fast Shared Response Model (FastSRM)

The implementation is based on the following publications:

.. [Richard2019] "Fast Shared Response Model for fMRI data"
    H. Richard, L. Martin, A. Pinho, J. Pillow, B. Thirion, 2019
    https://arxiv.org/pdf/1909.12537.pdf
"""

# Author: Hugo Richard

import hashlib
import logging
import os

import numpy as np
import scipy
from joblib import Parallel, delayed

from brainiak.funcalign.srm import DetSRM
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.exceptions import NotFittedError
import uuid

__all__ = [
    "FastSRM",
]

logger = logging.getLogger(__name__)


def get_shape(path):
    """Get shape of saved np array
    Parameters
    ----------
    path: str
        path to np array
    """
    f = open(path, "rb")
    version = np.lib.format.read_magic(f)
    shape, fortran_order, dtype = np.lib.format._read_array_header(f, version)
    f.close()
    return shape


def safe_load(data):
    """If data is an array returns data else returns np.load(data)"""
    if isinstance(data, np.ndarray):
        return data
    else:
        return np.load(data)


def safe_encode(img):
    if isinstance(img, np.ndarray):
        name = hashlib.md5(img.tostring()).hexdigest()
    else:
        name = hashlib.md5(img.encode()).hexdigest()
    return name


def assert_non_empty_list(input_list, list_name):
    """
    Check that input list is not empty
    Parameters
    ----------
    input_list: list
    list_name: str
        Name of the list
    """
    if len(input_list) == 0:
        raise ValueError("%s is a list of length 0 which is not valid" %
                         list_name)


def assert_array_2axis(array, name_array):
    """Check that input is an np array with 2 axes

    Parameters
    ----------
    array: np array
    name_array: str
        Name of the array
    """

    if not isinstance(array, np.ndarray):
        raise ValueError("%s should be of type "
                         "np.ndarray but is of type %s" %
                         (name_array, type(array)))

    if len(array.shape) != 2:
        raise ValueError("%s must have exactly 2 axes "
                         "but has %i axes" % (name_array, len(array.shape)))


def assert_valid_index(indexes, max_value, name_indexes):
    """
    Check that indexes are between 0 and max_value and number
    of indexes is less than max_value
    """
    for i, ind_i in enumerate(indexes):
        if ind_i < 0 or ind_i >= max_value:
            raise ValueError("Index %i of %s has value %i "
                             "whereas value should be between 0 and %i" %
                             (i, name_indexes, ind_i, max_value - 1))


def _check_imgs_list(imgs):
    """
    Checks that imgs is a non empty list of elements of the same type

    Parameters
    ----------

    imgs : list
    """
    # Check the list is non empty
    assert_non_empty_list(imgs, "imgs")

    # Check that all input have same type
    for i in range(len(imgs)):
        if not isinstance(imgs[i], type(imgs[0])):
            raise ValueError("imgs[%i] has type %s whereas "
                             "imgs[%i] has type %s. "
                             "This is inconsistent." %
                             (i, type(imgs[i]), 0, type(imgs[0])))


def _check_imgs_list_list(imgs):
    """
    Check input images if they are list of list of arrays

    Parameters
    ----------

    imgs : list of list of array of shape [n_voxels, n_components]
            imgs is a list of list of arrays where element i, j of
            the array is a numpy array of shape [n_voxels, n_timeframes] that
            contains the data of subject i collected during session j.
            n_timeframes and n_voxels are assumed to be the same across
            subjects
            n_timeframes can vary across sessions
            Each voxel's timecourse is assumed to have mean 0 and variance 1

    Returns
    -------
    shapes: array
        Shape of input images
    """
    n_subjects = len(imgs)

    # Check that the number of session is not 0
    assert_non_empty_list(imgs[0], "imgs[%i]" % 0)

    # Check that the number of sessions is the same for all subjects
    n_sessions = None
    for i in range(len(imgs)):
        if n_sessions is None:
            n_sessions = len(imgs[i])
        if n_sessions != len(imgs[i]):
            raise ValueError("imgs[%i] has length %i whereas imgs[%i] "
                             "has length %i. All subjects should have "
                             "the same number of sessions." %
                             (i, len(imgs[i]), 0, len(imgs[0])))

    shapes = np.zeros((n_subjects, n_sessions, 2))
    # Run array-level checks
    for i in range(len(imgs)):
        for j in range(len(imgs[i])):
            assert_array_2axis(imgs[i][j], "imgs[%i][%i]" % (i, j))
            shapes[i, j, :] = imgs[i][j].shape

    return shapes


def _check_imgs_list_array(imgs):
    """
    Check input images if they are list of arrays.
    In this case returned images are a list of list of arrays
    where element i,j of the array is a numpy array of
    shape [n_voxels, n_timeframes] that contains the data of subject i
    collected during session j.

    Parameters
    ----------

    imgs : array of str, shape=[n_subjects, n_sessions]
            imgs is a list of arrays where element i of the array is
            a numpy array of shape [n_voxels, n_timeframes] that contains the
            data of subject i (number of sessions is implicitly 1)
            n_timeframes and n_voxels are assumed to be the same across
            subjects
            n_timeframes can vary across sessions
            Each voxel's timecourse is assumed to have mean 0 and variance 1

    Returns
    -------
    shapes: array
        Shape of input images
    new_imgs: list of list of array of shape [n_voxels, n_components]
    """
    n_subjects = len(imgs)
    n_sessions = 1
    shapes = np.zeros((n_subjects, n_sessions, 2))
    new_imgs = []
    for i in range(len(imgs)):
        assert_array_2axis(imgs[i], "imgs[%i]" % i)
        shapes[i, 0, :] = imgs[i].shape
        new_imgs.append([imgs[i]])

    return new_imgs, shapes


def _check_imgs_array(imgs):
    """Check input image if it is an array

    Parameters
    ----------
    imgs : array of str, shape=[n_subjects, n_sessions]
            Element i, j of the array is a path to the data of subject i
            collected during session j.
            Data are loaded with numpy.load and expected
            shape is [n_voxels, n_timeframes]
            n_timeframes and n_voxels are assumed to be the same across
            subjects
            n_timeframes can vary across sessions
            Each voxel's timecourse is assumed to have mean 0 and variance 1

    Returns
    -------
    shapes : array
        Shape of input images
    """
    assert_array_2axis(imgs, "imgs")
    n_subjects, n_sessions = imgs.shape

    shapes = np.zeros((n_subjects, n_sessions, 2))
    for i in range(n_subjects):
        for j in range(n_sessions):
            if not (isinstance(imgs[i, j], str) or isinstance(
                    imgs[i, j], np.str_) or isinstance(imgs[i, j], str)):
                raise ValueError("imgs[%i, %i] is stored using "
                                 "type %s which is not a str" %
                                 (i, j, type(imgs[i, j])))
            shapes[i, j, :] = get_shape(imgs[i, j])
    return shapes


def _check_shapes_components(n_components, n_timeframes):
    """Check that n_timeframes is greater than number of components"""


def _check_shapes_atlas_compatibility(n_voxels,
                                      n_timeframes,
                                      n_components=None,
                                      atlas_shape=None):
    if n_components is not None:
        if np.sum(n_timeframes) < n_components:
            raise ValueError("Total number of timeframes is shorter than "
                             "number of components (%i < %i)" %
                             (np.sum(n_timeframes), n_components))

    if atlas_shape is not None:
        n_supervoxels, n_atlas_voxels = atlas_shape
        if n_atlas_voxels != n_voxels:
            raise ValueError(
                "Number of voxels in the atlas is not the same "
                "as the number of voxels in input data (%i != %i)" %
                (n_atlas_voxels, n_voxels))


def _check_shapes(shapes,
                  n_components=None,
                  atlas_shape=None,
                  ignore_nsubjects=False):
    """Check that number of voxels is the same for each subjects. Number of
    timeframes can vary between sessions but must be consistent across
    subjects

    Parameters
    ----------
    shapes : array of shape (n_subjects, n_sessions, 2)
        Array of shapes of input images
    """
    n_subjects, n_sessions, _ = shapes.shape

    if n_subjects <= 1 and not ignore_nsubjects:
        raise ValueError("The number of subjects should be greater than 1")

    n_timeframes_list = [None] * n_sessions
    n_voxels = None
    for n in range(n_subjects):
        for m in range(n_sessions):
            if n_timeframes_list[m] is None:
                n_timeframes_list[m] = shapes[n, m, 1]

            if n_voxels is None:
                n_voxels = shapes[m, n, 0]

            if n_timeframes_list[m] != shapes[n, m, 1]:
                raise ValueError("Subject %i Session %i does not have the "
                                 "same number of timeframes "
                                 "as Subject %i Session %i" % (n, m, 0, m))

            if n_voxels != shapes[n, m, 0]:
                raise ValueError("Subject %i Session %i"
                                 " does not have the same number of voxels as "
                                 "Subject %i Session %i." % (n, m, 0, 0))

    _check_shapes_atlas_compatibility(n_voxels, np.sum(n_timeframes_list),
                                      n_components, atlas_shape)


def check_atlas(atlas, n_components=None):
    """ Check input atlas

    Parameters
    ----------
    atlas :  array, shape=[n_supervoxels, n_voxels] or array, shape=[n_voxels]
     or str or None
        Probabilistic or deterministic atlas on which to project the data
        Deterministic atlas is an array of shape [n_voxels,] where values
        range from 1 to n_supervoxels. Voxels labelled 0 will be ignored.
        If atlas is a str the corresponding array is loaded with numpy.load
        and expected shape is (n_voxels,) for a deterministic atlas and
        (n_supervoxels, n_voxels) for a probabilistic atlas.

    n_components : int
        Number of timecourses of the shared coordinates

    Returns
    -------
    shape : array or None
        atlas shape
    """
    if atlas is None:
        return None

    if not (isinstance(atlas, np.ndarray) or isinstance(atlas, str)
            or isinstance(atlas, np.str_) or isinstance(atlas, str)):
        raise ValueError("Atlas is stored using "
                         "type %s which is neither np.ndarray or str" %
                         type(atlas))

    if isinstance(atlas, np.ndarray):
        shape = atlas.shape
    else:
        shape = get_shape(atlas)

    if len(shape) == 1:
        # We have a deterministic atlas
        atlas_array = safe_load(atlas)
        n_voxels = atlas_array.shape[0]
        n_supervoxels = len(np.unique(atlas_array)) - 1
        shape = (n_supervoxels, n_voxels)
    elif len(shape) != 2:
        raise ValueError(
            "Atlas has %i axes. It should have either 1 or 2 axes." %
            len(shape))

    n_supervoxels, n_voxels = shape

    if n_supervoxels > n_voxels:
        raise ValueError("Number of regions in the atlas is bigger than "
                         "the number of voxels (%i > %i)" %
                         (n_supervoxels, n_voxels))

    if n_components is not None:
        if n_supervoxels < n_components:
            raise ValueError("Number of regions in the atlas is "
                             "lower than the number of components "
                             "(%i < %i)" % (n_supervoxels, n_components))
    return shape


def check_imgs(imgs,
               n_components=None,
               atlas_shape=None,
               ignore_nsubjects=False):
    """
    Check input images

    Parameters
    ----------

    imgs : array of str, shape=[n_subjects, n_sessions]
            Element i, j of the array is a path to the data of subject i
            collected during session j.
            Data are loaded with numpy.load and expected
            shape is [n_voxels, n_timeframes]
            n_timeframes and n_voxels are assumed to be the same across
            subjects
            n_timeframes can vary across sessions
            Each voxel's timecourse is assumed to have mean 0 and variance 1

            imgs can also be a list of list of arrays where element i, j of
            the array is a numpy array of shape [n_voxels, n_timeframes] that
            contains the data of subject i collected during session j.

            imgs can also be a list of arrays where element i of the array is
            a numpy array of shape [n_voxels, n_timeframes] that contains the
            data of subject i (number of sessions is implicitly 1)

    Returns
    -------
    reshaped_input: bool
        True if input had to be reshaped to match the
        n_subjects, n_sessions input
    new_imgs: list of list of array or np array
        input imgs reshaped if it is a list of arrays so that it becomes a
        list of list of arrays
    shapes: array
        Shape of input images
    """
    reshaped_input = False
    new_imgs = imgs
    if isinstance(imgs, list):
        _check_imgs_list(imgs)
        if isinstance(imgs[0], list):
            shapes = _check_imgs_list_list(imgs)
        elif isinstance(imgs[0], np.ndarray):
            new_imgs, shapes = _check_imgs_list_array(imgs)
            reshaped_input = True
        else:
            raise ValueError(
                "Since imgs is a list, it should be a list of list "
                "of arrays or a list of arrays but imgs[0] has type %s" %
                type(imgs[0]))
    elif isinstance(imgs, np.ndarray):
        shapes = _check_imgs_array(imgs)
    else:
        raise ValueError(
            "Input imgs should either be a list or an array but has type %s" %
            type(imgs))

    _check_shapes(shapes, n_components, atlas_shape, ignore_nsubjects)

    return reshaped_input, new_imgs, shapes


def check_indexes(indexes, name):
    if not (indexes is None or isinstance(indexes, list)
            or isinstance(indexes, np.ndarray)):
        raise ValueError(
            "%s should be either a list, an array or None but received type %s"
            % (name, type(indexes)))


def _check_shared_response_list_of_list(shared_response, n_components,
                                        input_shapes):

    # Check that shared_response is indeed a list of list of arrays
    n_subjects = len(shared_response)
    n_sessions = None
    for i in range(len(shared_response)):
        if not isinstance(shared_response[i], list):
            raise ValueError("shared_response[0] is a list but "
                             "shared_response[%i] is not a list "
                             "this is incompatible." % i)
        assert_non_empty_list(shared_response[i], "shared_response[%i]" % i)
        if n_sessions is None:
            n_sessions = len(shared_response[i])
        elif n_sessions != len(shared_response[i]):
            raise ValueError(
                "shared_response[%i] has len %i whereas "
                "shared_response[0] has len %i. They should "
                "have same length" %
                (i, len(shared_response[i]), len(shared_response[0])))
        for j in range(len(shared_response[i])):
            assert_array_2axis(shared_response[i][j],
                               "shared_response[%i][%i]" % (i, j))

    return _check_shared_response_list_sessions([
        np.mean([shared_response[i][j] for i in range(n_subjects)], axis=0)
        for j in range(n_sessions)
    ], n_components, input_shapes)


def _check_shared_response_list_sessions(shared_response, n_components,
                                         input_shapes):
    for j in range(len(shared_response)):
        assert_array_2axis(shared_response[j], "shared_response[%i]" % j)
        if input_shapes is not None:
            if shared_response[j].shape[1] != input_shapes[0][j][1]:
                raise ValueError(
                    "Number of timeframes in input images during "
                    "session %i does not match the number of "
                    "timeframes during session %i "
                    "of shared_response (%i != %i)" %
                    (j, j, shared_response[j].shape[1], input_shapes[0, j, 1]))
        if n_components is not None:
            if shared_response[j].shape[0] != n_components:
                raise ValueError(
                    "Number of components in "
                    "shared_response during session %i is "
                    "different than "
                    "the number of components of the model (%i != %i)" %
                    (j, shared_response[j].shape[0], n_components))
    return shared_response


def _check_shared_response_list_subjects(shared_response, n_components,
                                         input_shapes):
    for i in range(len(shared_response)):
        assert_array_2axis(shared_response[i], "shared_response[%i]" % i)

    return _check_shared_response_array(np.mean(shared_response, axis=0),
                                        n_components, input_shapes)


def _check_shared_response_array(shared_response, n_components, input_shapes):
    assert_array_2axis(shared_response, "shared_response")
    if input_shapes is None:
        new_input_shapes = None
    else:
        n_subjects, n_sessions, _ = input_shapes.shape
        new_input_shapes = np.zeros((n_subjects, 1, 2))
        new_input_shapes[:, 0, 0] = input_shapes[:, 0, 0]
        new_input_shapes[:, 0, 1] = np.sum(input_shapes[:, :, 1], axis=1)
    return _check_shared_response_list_sessions([shared_response],
                                                n_components, new_input_shapes)


def check_shared_response(shared_response,
                          aggregate="mean",
                          n_components=None,
                          input_shapes=None):
    """
    Check that shared response has valid input and turn it into
    a session-wise shared response

    Returns
    -------
    added_session: bool
        True if an artificial sessions was added to match the list of
        session input type for shared_response
    reshaped_shared_response: list of arrays
        shared response (reshaped to match the list of session input)
    """
    # Depending on aggregate and shape of input we infer what to do
    if isinstance(shared_response, list):
        assert_non_empty_list(shared_response, "shared_response")
        if isinstance(shared_response[0], list):
            if aggregate == "mean":
                raise ValueError("self.aggregate has value 'mean' but "
                                 "shared response is a list of list. This is "
                                 "incompatible")
            return False, _check_shared_response_list_of_list(
                shared_response, n_components, input_shapes)
        elif isinstance(shared_response[0], np.ndarray):
            if aggregate == "mean":
                return False, _check_shared_response_list_sessions(
                    shared_response, n_components, input_shapes)
            else:
                return True, _check_shared_response_list_subjects(
                    shared_response, n_components, input_shapes)
        else:
            raise ValueError("shared_response is a list but "
                             "shared_response[0] is neither a list "
                             "or an array. This is invalid.")
    elif isinstance(shared_response, np.ndarray):
        return True, _check_shared_response_array(shared_response,
                                                  n_components, input_shapes)
    else:
        raise ValueError("shared_response should be either "
                         "a list or an array but is of type %s" %
                         type(shared_response))


def create_temp_dir(temp_dir):
    """
    This check whether temp_dir exists and creates dir otherwise
    """
    if temp_dir is None:
        return None

    if not os.path.exists(temp_dir):
        os.makedirs(temp_dir)
    else:
        raise ValueError("Path %s already exists. "
                         "When a model is used, filesystem should be cleaned "
                         "by using the .clean() method" % temp_dir)


def reduce_data_single(subject_index,
                       session_index,
                       img,
                       atlas=None,
                       inv_atlas=None,
                       low_ram=False,
                       temp_dir=None):
    """Reduce data using given atlas

    Parameters
    ----------
    subject_index : int

    session_index : int

    img : str or array
        path to data.
        Data are loaded with numpy.load and expected shape is
         (n_voxels, n_timeframes)
        n_timeframes and n_voxels are assumed to be the same across subjects
        n_timeframes can vary across sessions
        Each voxel's timecourse is assumed to have mean 0 and variance 1

        img can also be an array of shape (n_voxels, n_timeframes)

    atlas :  array, shape=[n_supervoxels, n_voxels] or [n_voxels] or None
        Probabilistic or deterministic atlas on which to project the data
        Deterministic atlas is an array of shape [n_voxels,] where values
        range from 1 to n_supervoxels. Voxels labelled 0 will be ignored.

    inv_atlas : array, shape=[n_voxels, n_supervoxels] or None
        Pseudo inverse of the atlas (only for probabilistic atlases)

    temp_dir : str or None
        path to dir where temporary results are stored
        if None temporary results will be stored in memory. This
        can results in memory errors when the number of subjects
        and / or sessions is large

    low_ram : bool
        if True and temp_dir is not None, reduced_data will be saved on disk
        this increases the number of IO but reduces memory complexity when the
         number
        of subject and number of sessions are large

    Returns
    -------

    reduced_data : array, shape=[n_timeframes, n_supervoxels]
        reduced data
    """
    # Here we return to the conventions of the paper
    data = safe_load(img).T

    n_timeframes, n_voxels = data.shape

    # Here we check that input is normalized
    if (np.max(np.abs(np.mean(data, axis=0))) > 1e-6
            or np.max(np.abs(np.var(data, axis=0) - 1))) > 1e-6:
        ValueError("Data in imgs[%i, %i] does not have 0 mean and unit \
        variance. If you are using NiftiMasker to mask your data \
        (nilearn) please use standardize=True." %
                   (subject_index, session_index))

    if inv_atlas is None and atlas is not None:
        atlas_values = np.unique(atlas)
        if 0 in atlas_values:
            atlas_values = atlas_values[1:]

        reduced_data = np.array(
            [np.mean(data[:, atlas == c], axis=1) for c in atlas_values]).T
    elif inv_atlas is not None and atlas is None:
        # this means that it is a probabilistic atlas
        reduced_data = data.dot(inv_atlas)
    else:
        reduced_data = data

    if low_ram:
        name = safe_encode(img)
        path = os.path.join(temp_dir, "reduced_data_" + name)
        np.save(path, reduced_data)
        return path + ".npy"
    else:
        return reduced_data


def reduce_data(imgs, atlas, n_jobs=1, low_ram=False, temp_dir=None):
    """Reduce data using given atlas.
    Work done in parallel across subjects.

    Parameters
    ----------

    imgs : array of str, shape=[n_subjects, n_sessions]
        Element i, j of the array is a path to the data of subject i
        collected during session j.
        Data are loaded with numpy.load and expected shape is
        [n_timeframes, n_voxels]
        n_timeframes and n_voxels are assumed to be the same across subjects
        n_timeframes can vary across sessions
        Each voxel's timecourse is assumed to have mean 0 and variance 1

        imgs can also be a list of list of arrays where element i, j of
        the array is a numpy array of shape [n_voxels, n_timeframes] that
        contains the data of subject i collected during session j.

        imgs can also be a list of arrays where element i of the array is
        a numpy array of shape [n_voxels, n_timeframes] that contains the
        data of subject i (number of sessions is implicitly 1)

    atlas :  array, shape=[n_supervoxels, n_voxels] or array, shape=[n_voxels]
        or None
        Probabilistic or deterministic atlas on which to project the data
        Deterministic atlas is an array of shape [n_voxels,] where values
        range from 1 to n_supervoxels. Voxels labelled 0 will be ignored.

    n_jobs : integer, optional, default=1
        The number of CPUs to use to do the computation.
         -1 means all CPUs, -2 all CPUs but one, and so on.

    temp_dir : str or None
        path to dir where temporary results are stored
        if None temporary results will be stored in memory. This
        can results in memory errors when the number of subjects
        and / or sessions is large

    low_ram : bool
        if True and temp_dir is not None, reduced_data will be saved on disk
        this increases the number of IO but reduces memory complexity when
         the number of subject and/or sessions is large

    Returns
    -------

    reduced_data_list : array of str, shape=[n_subjects, n_sessions]
    or array, shape=[n_subjects, n_sessions, n_timeframes, n_supervoxels]
        Element i, j of the array is a path to the data of subject i collected
         during session j.
        Data are loaded with numpy.load and expected shape is
        [n_timeframes, n_supervoxels]
        or Element i, j of the array is the data in array of
        shape=[n_timeframes, n_supervoxels]
        n_timeframes and n_supervoxels
         are assumed to be the same across subjects
        n_timeframes can vary across sessions
        Each voxel's timecourse is assumed to have mean 0 and variance 1
    """
    if atlas is None:
        A = None
        A_inv = None
    else:
        loaded_atlas = safe_load(atlas)
        if len(loaded_atlas.shape) == 2:
            A = None
            A_inv = loaded_atlas.T.dot(
                np.linalg.inv(loaded_atlas.dot(loaded_atlas.T)))
        else:
            A = loaded_atlas
            A_inv = None

    n_subjects = len(imgs)
    n_sessions = len(imgs[0])

    reduced_data_list = Parallel(n_jobs=n_jobs)(
        delayed(reduce_data_single)(i,
                                    j,
                                    imgs[i][j],
                                    atlas=A,
                                    inv_atlas=A_inv,
                                    low_ram=low_ram,
                                    temp_dir=temp_dir)
        for i in range(n_subjects) for j in range(n_sessions))

    # Check if reduced_data_list is ragged, in this case, we need to create a
    # numpy array with dtype=object, numpy no longer does this automatically
    try:
        reduced_data_list = np.array(reduced_data_list)
    except ValueError as ex:
        if "setting an array element with a sequence" in str(ex):
            reduced_data_list = np.array(reduced_data_list, dtype=object)
        else:
            raise ex

    if low_ram:
        reduced_data_list = np.reshape(reduced_data_list,
                                       (n_subjects, n_sessions))
    else:
        if len(np.array(reduced_data_list).shape) == 1:
            reduced_data_list = np.reshape(reduced_data_list,
                                           (n_subjects, n_sessions))
        else:
            n_timeframes, n_supervoxels = np.array(reduced_data_list).shape[1:]
            reduced_data_list = np.reshape(
                reduced_data_list,
                (n_subjects, n_sessions, n_timeframes, n_supervoxels))

    return reduced_data_list


def _reduced_space_compute_shared_response(reduced_data_list,
                                           reduced_basis_list,
                                           n_components=50):
    """Compute shared response with basis fixed in reduced space

    Parameters
    ----------

    reduced_data_list : array of str, shape=[n_subjects, n_sessions]
    or array, shape=[n_subjects, n_sessions, n_timeframes, n_supervoxels]
        Element i, j of the array is a path to the data of subject i
        collected during session j.
        Data are loaded with numpy.load and expected shape is
        [n_timeframes, n_supervoxels]
        or Element i, j of the array is the data in array of
        shape=[n_timeframes, n_supervoxels]
        n_timeframes and n_supervoxels are
        assumed to be the same across subjects
        n_timeframes can vary across sessions
        Each voxel's timecourse is assumed to have mean 0 and variance 1

    reduced_basis_list : None or list of array, element i has
    shape=[n_components, n_supervoxels]
        each subject's reduced basis
        if None the basis will be generated on the fly

    n_components : int or None
        number of components

    Returns
    -------

    shared_response_list : list of array, element i has
    shape=[n_timeframes, n_components]
        shared response, element i is the shared response during session i

    """
    n_subjects, n_sessions = reduced_data_list.shape[:2]

    s = [None] * n_sessions

    # This is just to check that all subjects have same number of
    # timeframes in a given session
    for n in range(n_subjects):
        for m in range(n_sessions):
            data_nm = safe_load(reduced_data_list[n][m])
            n_timeframes, n_supervoxels = data_nm.shape

            if reduced_basis_list is None:
                reduced_basis_list = []
                for subject in range(n_subjects):
                    q = np.eye(n_components, n_supervoxels)
                    reduced_basis_list.append(q)

            basis_n = reduced_basis_list[n]
            if s[m] is None:
                s[m] = data_nm.dot(basis_n.T)
            else:
                s[m] = s[m] + data_nm.dot(basis_n.T)

    for m in range(n_sessions):
        s[m] = s[m] / float(n_subjects)

    return s


def _compute_and_save_corr_mat(img, shared_response, temp_dir):
    """computes correlation matrix and stores it

    Parameters
    ----------
    img : str
        path to data.
        Data are loaded with numpy.load and expected shape is
        [n_timeframes, n_voxels]
        n_timeframes and n_voxels are assumed to be the same across subjects
        n_timeframes can vary across sessions
        Each voxel's timecourse is assumed to have mean 0 and variance 1

    shared_response : array, shape=[n_timeframes, n_components]
        shared response
    """
    data = safe_load(img).T
    name = safe_encode(img)
    path = os.path.join(temp_dir, "corr_mat_" + name)
    np.save(path, shared_response.T.dot(data))


def _compute_and_save_subject_basis(subject_number, sessions, temp_dir):
    """computes correlation matrix for all sessions

    Parameters
    ----------

    subject_number: int
        Number that identifies the subject. Basis will be stored in
         [temp_dir]/basis_[subject_number].npy

    sessions : array of str
        Element i of the array is a path to the data collected during
        session i.
        Data are loaded with numpy.load and expected shape is
        [n_timeframes, n_voxels]
        n_timeframes and n_voxels are assumed to be the same across subjects
        n_timeframes can vary across sessions
        Each voxel's timecourse is assumed to have mean 0 and variance

    temp_dir : str or None
        path to dir where temporary results are stored
        if None temporary results will be stored in memory. This
        can results in memory errors when the number of subjects
        and / or sessions is large

    Returns
    -------

    basis: array, shape=[n_component, n_voxels] or str
        basis of subject [subject_number] or path to this basis
    """
    corr_mat = None
    for session in sessions:
        name = safe_encode(session)
        path = os.path.join(temp_dir, "corr_mat_" + name + ".npy")
        if corr_mat is None:
            corr_mat = np.load(path)
        else:
            corr_mat += np.load(path)
        os.remove(path)
    basis_i = _compute_subject_basis(corr_mat)
    path = os.path.join(temp_dir, "basis_%i" % subject_number)
    np.save(path, basis_i)
    return path + ".npy"


def _compute_subject_basis(corr_mat):
    """From correlation matrix between shared response and subject data,
    Finds subject's basis

    Parameters
    ----------

    corr_mat: array, shape=[n_component, n_voxels]
    or shape=[n_components, n_supervoxels]
        correlation matrix between shared response and subject data or
         subject reduced data
        element k, v is given by S.T.dot(X_i) where S is the shared response
         and X_i the data of subject i.

    Returns
    -------

    basis: array, shape=[n_components, n_voxels]
    or shape=[n_components, n_supervoxels]
        basis of subject or reduced_basis of subject
    """
    # The perturbation is only here to be
    # consistent with current implementation
    # of DetSRM.
    perturbation = np.zeros(corr_mat.shape)
    np.fill_diagonal(perturbation, 0.001)
    U, _, V = scipy.linalg.svd(corr_mat + perturbation, full_matrices=False)
    return U.dot(V)


def fast_srm(reduced_data_list,
             n_iter=10,
             n_components=None,
             low_ram=False,
             seed=0):
    """Computes shared response and basis in reduced space

    Parameters
    ----------

    reduced_data_list : array, shape=[n_subjects, n_sessions]
    or array, shape=[n_subjects, n_sessions, n_timeframes, n_supervoxels]
        Element i, j of the array is a path to the data of subject i
        collected during session j.
        Data are loaded with numpy.load and expected
        shape is [n_timeframes, n_supervoxels]
        or Element i, j of the array is the data in array of
        shape=[n_timeframes, n_supervoxels]
        n_timeframes and n_supervoxels are
         assumed to be the same across subjects
        n_timeframes can vary across sessions
        Each voxel's timecourse is assumed to have mean 0 and variance 1

    n_iter : int
        Number of iterations performed

    n_components : int or None
        number of components

    Returns
    -------

    shared_response_list : list of array, element i has
     shape=[n_timeframes, n_components]
        shared response, element i is the shared response during session i
    """
    if low_ram:
        return lowram_srm(reduced_data_list, n_iter, n_components)
    else:
        # We need to switch data to DetSRM format
        # Indeed in DetSRM all sessions are concatenated.
        # Whereas in FastSRM multiple sessions are supported.

        n_subjects, n_sessions = reduced_data_list.shape[:2]
        # We store the correspondence between timeframes and session
        timeframes_slices = []
        current_j = 0
        for j in range(n_sessions):
            timeframes_slices.append(
                slice(current_j, current_j + len(reduced_data_list[0, j])))
            current_j += len(reduced_data_list[0][j])
        # Now we can concatenate everything
        X = [
            np.concatenate(reduced_data_list[i], axis=0).T
            for i in range(n_subjects)
        ]

        srm = DetSRM(n_iter=n_iter, features=n_components, rand_seed=seed)
        srm.fit(X)

        # SRM gives a list of data projected in shared space
        # we get the shared response by averaging those
        concatenated_s = np.mean(srm.transform(X), axis=0).T

        # Let us return the shared response sliced by sessions
        return [concatenated_s[i] for i in timeframes_slices]


def lowram_srm(reduced_data_list, n_iter=10, n_components=None):
    """Computes shared response and basis in reduced space

    Parameters
    ----------

    reduced_data_list : array of str, shape=[n_subjects, n_sessions]
        Element i, j of the array is a path to the data of subject i
        collected during session j.
        Data are loaded with numpy.load and expected
        shape is [n_timeframes, n_supervoxels]
        n_timeframes and n_supervoxels are
         assumed to be the same across subjects
        n_timeframes can vary across sessions
        Each voxel's timecourse is assumed to have mean 0 and variance 1

    n_iter : int
        Number of iterations performed

    n_components : int or None
        number of components

    Returns
    -------

    shared_response_list : list of array, element i has
     shape=[n_timeframes, n_components]
        shared response, element i is the shared response during session i
    """

    n_subjects, n_sessions = reduced_data_list.shape[:2]
    shared_response = _reduced_space_compute_shared_response(
        reduced_data_list, None, n_components)

    reduced_basis = [None] * n_subjects
    for _ in range(n_iter):
        for n in range(n_subjects):
            cov = None
            for m in range(n_sessions):
                data_nm = np.load(reduced_data_list[n, m])
                if cov is None:
                    cov = shared_response[m].T.dot(data_nm)
                else:
                    cov += shared_response[m].T.dot(data_nm)
            reduced_basis[n] = _compute_subject_basis(cov)

        shared_response = _reduced_space_compute_shared_response(
            reduced_data_list, reduced_basis, n_components)

    return shared_response


def _compute_basis_subject_online(sessions, shared_response_list):
    """Computes subject's basis with shared response fixed

    Parameters
    ----------

    sessions : array of str
        Element i of the array is a path to the data
        collected during session i.
        Data are loaded with numpy.load and expected shape is
         [n_timeframes, n_voxels]
        n_timeframes and n_voxels are assumed to be the same across subjects
        n_timeframes can vary across sessions
        Each voxel's timecourse is assumed to have mean 0 and variance 1

    shared_response_list : list of array, element i has
    shape=[n_timeframes, n_components]
        shared response, element i is the shared response during session i

    Returns
    -------

    basis: array, shape=[n_components, n_voxels]
        basis
    """

    basis_i = None
    i = 0
    for session in sessions:
        data = safe_load(session).T
        if basis_i is None:
            basis_i = shared_response_list[i].T.dot(data)
        else:
            basis_i += shared_response_list[i].T.dot(data)
        i += 1
        del data
    return _compute_subject_basis(basis_i)


def _compute_shared_response_online_single(subjects, basis_list, temp_dir,
                                           subjects_indexes, aggregate):
    """Computes shared response during one session with basis fixed

    Parameters
    ----------

    subjects : array of str
        Element i of the array is a path to the data of subject i.
        Data are loaded with numpy.load and expected shape is
         [n_timeframes, n_voxels]
        n_timeframes and n_voxels are assumed to be the same across subjects
        n_timeframes can vary across sessions
        Each voxel's timecourse is assumed to have mean 0 and variance 1

    basis_list : None or list of array, element i has
    shape=[n_components, n_voxels]
        basis of all subjects, element i is the basis of subject i

    temp_dir : None or str
        path to basis folder where file basis_%i.npy contains the basis of
        subject i

    subjects_indexes : list of int or None
        list of indexes corresponding to the subjects to use to compute
        shared response

    aggregate: str or None, default="mean"
        if "mean": returns the mean shared response S from all subjects
        if None: returns the subject-specific response in shared space S_i

    Returns
    -------

    shared_response : array, shape=[n_timeframes, n_components] or list
        shared response
    """
    n = 0
    if aggregate == "mean":
        shared_response = None
    if aggregate is None:
        shared_response = []

    for k, i in enumerate(subjects_indexes):
        subject = subjects[k]
        # Transpose to be consistent with paper
        data = safe_load(subject).T
        if temp_dir is None:
            basis_i = basis_list[i]
        else:
            basis_i = np.load(os.path.join(temp_dir, "basis_%i.npy" % i))

        if aggregate == "mean":
            if shared_response is None:
                shared_response = data.dot(basis_i.T)
            else:
                shared_response += data.dot(basis_i.T)
            n += 1

        if aggregate is None:
            shared_response.append(data.dot(basis_i.T))

    if aggregate is None:
        return shared_response

    if aggregate == "mean":
        return shared_response / float(n)


def _compute_shared_response_online(imgs, basis_list, temp_dir, n_jobs,
                                    subjects_indexes, aggregate):
    """Computes shared response with basis fixed

    Parameters
    ----------

    imgs : array of str, shape=[n_subjects, n_sessions]
        Element i, j of the array is a path to the data of subject i
        collected during session j.
        Data are loaded with numpy.load and expected shape is
        [n_timeframes, n_voxels]
        n_timeframes and n_voxels are assumed to be the same across subjects
        n_timeframes can vary across sessions
        Each voxel's timecourse is assumed to have mean 0 and variance 1

        imgs can also be a list of list of arrays where element i, j of
        the array is a numpy array of shape [n_voxels, n_timeframes] that
        contains the data of subject i collected during session j.

    basis_list : None or list of array, element i has
    shape=[n_components, n_voxels]
        basis of all subjects, element i is the basis of subject i

    temp_dir : None or str
        path to basis folder where file basis_%i.npy contains the basis of
        subject i

    n_jobs : integer, optional, default=1
            The number of CPUs to use to do the computation.
             -1 means all CPUs, -2 all CPUs but one, and so on.

    subjects_indexes : list or None
        list of indexes corresponding to the subjects to use to compute
        shared response

    aggregate: str or None, default="mean"
        if "mean": returns the mean shared response S from all subjects
        if None: returns the subject-specific response in shared space S_i

    Returns
    -------

    shared_response_list : list of array or list of list of array
        shared response, element i is the shared response during session i
        or element i, j is the shared response of subject i during session j
    """

    n_subjects = len(subjects_indexes)
    n_sessions = len(imgs[0])

    shared_response_list = Parallel(n_jobs=n_jobs)(
        delayed(_compute_shared_response_online_single)
        ([imgs[i][j] for i in range(n_subjects)], basis_list, temp_dir,
         subjects_indexes, aggregate) for j in range(n_sessions))

    if aggregate is None:
        shared_response_list = [[
            shared_response_list[j][i].T for j in range(n_sessions)
        ] for i in range(n_subjects)]

    if aggregate == "mean":
        shared_response_list = [
            shared_response_list[j].T for j in range(n_sessions)
        ]

    return shared_response_list


class FastSRM(BaseEstimator, TransformerMixin):
    """SRM decomposition using a very low amount of memory and \
computational power thanks to the use of an atlas \
as described in [Richard2019]_.

    Given multi-subject data, factorize it as a shared response S \
among all subjects and an orthogonal transform (basis) W per subject:

    .. math:: X_i \\approx W_i S, \\forall i=1 \\dots N

    Parameters
    ----------

    atlas :  array, shape=[n_supervoxels, n_voxels] or array,\
shape=[n_voxels] or str or None, default=None
        Probabilistic or deterministic atlas on which to project the data. \
Deterministic atlas is an array of shape [n_voxels,] \
where values range from 1 \
to n_supervoxels. Voxels labelled 0 will be ignored. If atlas is a str the \
corresponding array is loaded with numpy.load and expected shape \
is (n_voxels,) for a deterministic atlas and \
(n_supervoxels, n_voxels) for a probabilistic atlas.

    n_components : int
        Number of timecourses of the shared coordinates

    n_iter : int
        Number of iterations to perform

    temp_dir : str or None
        Path to dir where temporary results are stored. If None \
temporary results will be stored in memory. This can results in memory \
errors when the number of subjects and/or sessions is large

    low_ram : bool
        If True and temp_dir is not None, reduced_data will be saved on \
disk. This increases the number of IO but reduces memory complexity when \
the number of subject and/or sessions is large

    seed : int
        Seed used for random sampling.

    n_jobs : int, optional, default=1
        The number of CPUs to use to do the computation. \
-1 means all CPUs, -2 all CPUs but one, and so on.

    verbose : bool or "warn"
        If True, logs are enabled. If False, logs are disabled. \
If "warn" only warnings are printed.

    aggregate: str or None, default="mean"
        If "mean", shared_response is the mean shared response \
from all subjects. If None, shared_response contains all \
subject-specific responses in shared space

    Attributes
    ----------

    `basis_list`: list of array, element i has \
shape=[n_components, n_voxels] or list of str
        - if basis is a list of array, element i is the basis of subject i
        - if basis is a list of str, element i is the path to the basis \
of subject i that is loaded with np.load yielding an array of \
shape [n_components, n_voxels].

        Note that any call to the clean method erases this attribute

    Note
    -----

        **References:**
        H. Richard, L. Martin, A. Pinho, J. Pillow, B. Thirion, 2019: \
Fast shared response model for fMRI data (https://arxiv.org/pdf/1909.12537.pdf)

    """
    def __init__(self,
                 atlas=None,
                 n_components=20,
                 n_iter=100,
                 temp_dir=None,
                 low_ram=False,
                 seed=None,
                 n_jobs=1,
                 verbose="warn",
                 aggregate="mean"):

        self.seed = seed
        self.n_jobs = n_jobs
        self.verbose = verbose
        self.n_components = n_components
        self.n_iter = n_iter
        self.atlas = atlas

        if aggregate is not None and aggregate != "mean":
            raise ValueError("aggregate can have only value mean or None")

        self.aggregate = aggregate

        self.basis_list = None

        if temp_dir is None:
            if self.verbose == "warn" or self.verbose is True:
                logger.warning("temp_dir has value None. "
                               "All basis (spatial maps) and reconstructed "
                               "data will therefore be kept in memory."
                               "This can lead to memory errors when the "
                               "number of subjects "
                               "and/or sessions is large.")
            self.temp_dir = None
            self.low_ram = False

        if temp_dir is not None:
            self.temp_dir = os.path.join(temp_dir,
                                         "fastsrm" + str(uuid.uuid4()))
            self.low_ram = low_ram

    def clean(self):
        """This erases temporary files and basis_list attribute to \
free memory. This method should be called when fitted model \
is not needed anymore.
        """
        if self.temp_dir is not None:
            if os.path.exists(self.temp_dir):
                for root, dirs, files in os.walk(self.temp_dir, topdown=False):
                    for name in files:
                        os.remove(os.path.join(root, name))
                os.rmdir(self.temp_dir)

        if self.basis_list is not None:
            self.basis_list is None

    def fit(self, imgs):
        """Computes basis across subjects from input imgs

        Parameters
        ----------
        imgs : array of str, shape=[n_subjects, n_sessions] or \
list of list of arrays or list of arrays
            Element i, j of the array is a path to the data of subject i \
collected during session j. Data are loaded with numpy.load and expected \
shape is [n_voxels, n_timeframes] n_timeframes and n_voxels are assumed \
to be the same across subjects n_timeframes can vary across sessions. \
Each voxel's timecourse is assumed to have mean 0 and variance 1

            imgs can also be a list of list of arrays where element i, j \
of the array is a numpy array of shape [n_voxels, n_timeframes] \
that contains the data of subject i collected during session j.

            imgs can also be a list of arrays where element i \
of the array is a numpy array of shape [n_voxels, n_timeframes] \
that contains the data of subject i (number of sessions is implicitly 1)

        Returns
        -------
        self : object
           Returns the instance itself. Contains attributes listed \
at the object level.
        """
        atlas_shape = check_atlas(self.atlas, self.n_components)
        reshaped_input, imgs, shapes = check_imgs(
            imgs, n_components=self.n_components, atlas_shape=atlas_shape)
        self.clean()
        create_temp_dir(self.temp_dir)

        if self.verbose is True:
            logger.info("[FastSRM.fit] Reducing data")

        reduced_data = reduce_data(imgs,
                                   atlas=self.atlas,
                                   n_jobs=self.n_jobs,
                                   low_ram=self.low_ram,
                                   temp_dir=self.temp_dir)

        if self.verbose is True:
            logger.info("[FastSRM.fit] Finds shared "
                        "response using reduced data")
        shared_response_list = fast_srm(reduced_data,
                                        n_iter=self.n_iter,
                                        n_components=self.n_components,
                                        low_ram=self.low_ram,
                                        seed=self.seed)
        if self.verbose is True:
            logger.info("[FastSRM.fit] Finds basis using "
                        "full data and shared response")

        if self.n_jobs == 1:
            basis = []
            for i, sessions in enumerate(imgs):
                basis_i = _compute_basis_subject_online(
                    sessions, shared_response_list)
                if self.temp_dir is None:
                    basis.append(basis_i)
                else:
                    path = os.path.join(self.temp_dir, "basis_%i" % i)
                    np.save(path, basis_i)
                    basis.append(path + ".npy")
                del basis_i
        else:
            if self.temp_dir is None:
                basis = Parallel(n_jobs=self.n_jobs)(
                    delayed(_compute_basis_subject_online)(
                        sessions, shared_response_list) for sessions in imgs)
            else:
                Parallel(n_jobs=self.n_jobs)(
                    delayed(_compute_and_save_corr_mat)(
                        imgs[i][j], shared_response_list[j], self.temp_dir)
                    for j in range(len(imgs[0])) for i in range(len(imgs)))

                basis = Parallel(n_jobs=self.n_jobs)(
                    delayed(_compute_and_save_subject_basis)(i, sessions,
                                                             self.temp_dir)
                    for i, sessions in enumerate(imgs))

        self.basis_list = basis
        return self

    def fit_transform(self, imgs, subjects_indexes=None):
        """Computes basis across subjects and shared response from input imgs
        return shared response.

        Parameters
        ----------
        imgs : array of str, shape=[n_subjects, n_sessions] or \
list of list of arrays or list of arrays
            Element i, j of the array is a path to the data of subject i \
collected during session j. Data are loaded with numpy.load and expected \
shape is [n_voxels, n_timeframes] n_timeframes and n_voxels are assumed \
to be the same across subjects n_timeframes can vary across sessions. \
Each voxel's timecourse is assumed to have mean 0 and variance 1

            imgs can also be a list of list of arrays where element i, j \
of the array is a numpy array of shape [n_voxels, n_timeframes] \
that contains the data of subject i collected during session j.

            imgs can also be a list of arrays where element i \
of the array is a numpy array of shape [n_voxels, n_timeframes] \
that contains the data of subject i (number of sessions is implicitly 1)

        subjects_indexes : list or None:
            if None imgs[i] will be transformed using basis_list[i]. \
Otherwise imgs[i] will be transformed using basis_list[subjects_index[i]]

        Returns
        --------
        shared_response : list of arrays, list of list of arrays or array
            - if imgs is a list of array and self.aggregate="mean": shared \
response is an array of shape (n_components, n_timeframes)
            - if imgs is a list of array and self.aggregate=None: shared \
response is a list of array, element i is the projection of data of \
subject i in shared space.
            - if imgs is an array or a list of list of array and \
self.aggregate="mean": shared response is a list of array, \
element j is the shared response during session j
            - if imgs is an array or a list of list of array and \
self.aggregate=None: shared response is a list of list of array, \
element i, j is the projection of data of subject i collected \
during session j in shared space.
        """
        self.fit(imgs)
        return self.transform(imgs, subjects_indexes=subjects_indexes)

    def transform(self, imgs, subjects_indexes=None):
        """From data in imgs and basis from training data,
        computes shared response.

        Parameters
        ----------
        imgs : array of str, shape=[n_subjects, n_sessions] or \
list of list of arrays or list of arrays
            Element i, j of the array is a path to the data of subject i \
collected during session j. Data are loaded with numpy.load and expected \
shape is [n_voxels, n_timeframes] n_timeframes and n_voxels are assumed \
to be the same across subjects n_timeframes can vary across sessions. \
Each voxel's timecourse is assumed to have mean 0 and variance 1

            imgs can also be a list of list of arrays where element i, j \
of the array is a numpy array of shape [n_voxels, n_timeframes] \
that contains the data of subject i collected during session j.

            imgs can also be a list of arrays where element i \
of the array is a numpy array of shape [n_voxels, n_timeframes] \
that contains the data of subject i (number of sessions is implicitly 1)

        subjects_indexes : list or None:
            if None imgs[i] will be transformed using basis_list[i]. \
Otherwise imgs[i] will be transformed using basis[subjects_index[i]]

        Returns
        --------
        shared_response : list of arrays, list of list of arrays or array
            - if imgs is a list of array and self.aggregate="mean": shared \
response is an array of shape (n_components, n_timeframes)
            - if imgs is a list of array and self.aggregate=None: shared \
response is a list of array, element i is the projection of data of \
subject i in shared space.
            - if imgs is an array or a list of list of array and \
self.aggregate="mean": shared response is a list of array, \
element j is the shared response during session j
            - if imgs is an array or a list of list of array and \
self.aggregate=None: shared response is a list of list of array, \
element i, j is the projection of data of subject i collected \
during session j in shared space.
         """
        aggregate = self.aggregate
        if self.basis_list is None:
            raise NotFittedError("The model fit has not been run yet.")

        atlas_shape = check_atlas(self.atlas, self.n_components)
        reshaped_input, imgs, shapes = check_imgs(
            imgs,
            n_components=self.n_components,
            atlas_shape=atlas_shape,
            ignore_nsubjects=True)
        check_indexes(subjects_indexes, "subjects_indexes")
        if subjects_indexes is None:
            subjects_indexes = np.arange(len(imgs))
        else:
            subjects_indexes = np.array(subjects_indexes)

        # Transform specific checks
        if len(subjects_indexes) < len(imgs):
            raise ValueError("Input data imgs has len %i whereas "
                             "subject_indexes has len %i. "
                             "The number of basis used to compute "
                             "the shared response should be equal "
                             "to the number of subjects in imgs" %
                             (len(imgs), len(subjects_indexes)))

        assert_valid_index(subjects_indexes, len(self.basis_list),
                           "subjects_indexes")

        shared_response = _compute_shared_response_online(
            imgs, self.basis_list, self.temp_dir, self.n_jobs,
            subjects_indexes, aggregate)

        # If shared response has only 1 session we need to reshape it
        if reshaped_input:
            if aggregate == "mean":
                shared_response = shared_response[0]
            if aggregate is None:
                shared_response = [
                    shared_response[i][0] for i in range(len(subjects_indexes))
                ]

        return shared_response

    def inverse_transform(
            self,
            shared_response,
            subjects_indexes=None,
            sessions_indexes=None,
    ):
        """From shared response and basis from training data
        reconstruct subject's data

        Parameters
        ----------

        shared_response : list of arrays, list of list of arrays or array
            - if imgs is a list of array and self.aggregate="mean": shared \
response is an array of shape (n_components, n_timeframes)
            - if imgs is a list of array and self.aggregate=None: shared \
response is a list of array, element i is the projection of data of \
subject i in shared space.
            - if imgs is an array or a list of list of array and \
self.aggregate="mean": shared response is a list of array, \
element j is the shared response during session j
            - if imgs is an array or a list of list of array and \
self.aggregate=None: shared response is a list of list of array, \
element i, j is the projection of data of subject i collected \
during session j in shared space.

        subjects_indexes : list or None
            if None reconstructs data of all subjects used during train. \
Otherwise reconstructs data of subjects specified by subjects_indexes.

        sessions_indexes : list or None
            if None reconstructs data of all sessions. \
Otherwise uses reconstructs data of sessions specified by sessions_indexes.

        Returns
        -------
        reconstructed_data: list of list of arrays or list of arrays
            - if reconstructed_data is a list of list : element i, j is \
the reconstructed data for subject subjects_indexes[i] and \
session sessions_indexes[j] as an np array of shape n_voxels, \
n_timeframes
            - if reconstructed_data is a list : element i is the \
reconstructed data for subject \
subject_indexes[i] as an np array of shape n_voxels, n_timeframes

        """
        added_session, shared = check_shared_response(
            shared_response, self.aggregate, n_components=self.n_components)
        n_subjects = len(self.basis_list)
        n_sessions = len(shared)

        for j in range(n_sessions):
            assert_array_2axis(shared[j], "shared_response[%i]" % j)

        check_indexes(subjects_indexes, "subjects_indexes")
        check_indexes(sessions_indexes, "sessions_indexes")

        if subjects_indexes is None:
            subjects_indexes = np.arange(n_subjects)
        else:
            subjects_indexes = np.array(subjects_indexes)

        assert_valid_index(subjects_indexes, n_subjects, "subjects_indexes")

        if sessions_indexes is None:
            sessions_indexes = np.arange(len(shared))
        else:
            sessions_indexes = np.array(sessions_indexes)

        assert_valid_index(sessions_indexes, n_sessions, "sessions_indexes")

        data = []
        for i in subjects_indexes:
            data_ = []
            basis_i = safe_load(self.basis_list[i])
            if added_session:
                data.append(basis_i.T.dot(shared[0]))
            else:
                for j in sessions_indexes:
                    data_.append(basis_i.T.dot(shared[j]))
                data.append(data_)
        return data

    def add_subjects(self, imgs, shared_response):
        """ Add subjects to the current fit. Each new basis will be \
appended at the end of the list of basis (which can \
be accessed using self.basis)

        Parameters
        ----------

        imgs : array of str, shape=[n_subjects, n_sessions] or \
list of list of arrays or list of arrays
            Element i, j of the array is a path to the data of subject i \
collected during session j. Data are loaded with numpy.load and expected \
shape is [n_voxels, n_timeframes] n_timeframes and n_voxels are assumed \
to be the same across subjects n_timeframes can vary across sessions. \
Each voxel's timecourse is assumed to have mean 0 and variance 1

            imgs can also be a list of list of arrays where element i, j \
of the array is a numpy array of shape [n_voxels, n_timeframes] \
that contains the data of subject i collected during session j.

            imgs can also be a list of arrays where element i \
of the array is a numpy array of shape [n_voxels, n_timeframes] \
that contains the data of subject i (number of sessions is implicitly 1)

        shared_response : list of arrays, list of list of arrays or array
            - if imgs is a list of array and self.aggregate="mean": shared \
response is an array of shape (n_components, n_timeframes)
            - if imgs is a list of array and self.aggregate=None: shared \
response is a list of array, element i is the projection of data of \
subject i in shared space.
            - if imgs is an array or a list of list of array and \
self.aggregate="mean": shared response is a list of array, \
element j is the shared response during session j
            - if imgs is an array or a list of list of array and \
self.aggregate=None: shared response is a list of list of array, \
element i, j is the projection of data of subject i collected \
during session j in shared space.
        """
        atlas_shape = check_atlas(self.atlas, self.n_components)
        reshaped_input, imgs, shapes = check_imgs(
            imgs,
            n_components=self.n_components,
            atlas_shape=atlas_shape,
            ignore_nsubjects=True)

        _, shared_response_list = check_shared_response(
            shared_response,
            n_components=self.n_components,
            aggregate=self.aggregate,
            input_shapes=shapes)

        # we need to transpose shared_response_list to be consistent with
        # other functions
        shared_response_list = [
            shared_response_list[j].T for j in range(len(shared_response_list))
        ]

        if self.n_jobs == 1:
            basis = []
            for i, sessions in enumerate(imgs):
                basis_i = _compute_basis_subject_online(
                    sessions, shared_response_list)
                if self.temp_dir is None:
                    basis.append(basis_i)
                else:
                    path = os.path.join(
                        self.temp_dir, "basis_%i" % (len(self.basis_list) + i))
                    np.save(path, basis_i)
                    basis.append(path + ".npy")
                del basis_i
        else:
            if self.temp_dir is None:
                basis = Parallel(n_jobs=self.n_jobs)(
                    delayed(_compute_basis_subject_online)(
                        sessions, shared_response_list) for sessions in imgs)
            else:
                Parallel(n_jobs=self.n_jobs)(
                    delayed(_compute_and_save_corr_mat)(
                        imgs[i][j], shared_response_list[j], self.temp_dir)
                    for j in range(len(imgs[0])) for i in range(len(imgs)))

                basis = Parallel(n_jobs=self.n_jobs)(
                    delayed(_compute_and_save_subject_basis)(
                        len(self.basis_list) + i, sessions, self.temp_dir)
                    for i, sessions in enumerate(imgs))

        self.basis_list += basis
